from __future__ import annotations

import json

import pytest

from dagzoo.config import REQUEST_FILE_VERSION_V1, REQUEST_TASK_CLASSIFICATION, RequestFileConfig
from dagzoo.core.request_handoff import (
    REQUEST_HANDOFF_SCHEMA_NAME,
    REQUEST_HANDOFF_SCHEMA_VERSION,
    build_request_handoff_manifest,
    validate_request_handoff_manifest,
    write_request_handoff_manifest,
)


def _request_config(output_root: str) -> RequestFileConfig:
    return RequestFileConfig.from_dict(
        {
            "version": REQUEST_FILE_VERSION_V1,
            "task": REQUEST_TASK_CLASSIFICATION,
            "dataset_count": 2,
            "rows": 1024,
            "profile": "default",
            "output_root": output_root,
        }
    )


def test_build_request_handoff_manifest_is_versioned_and_valid(tmp_path) -> None:
    request = _request_config(output_root="requests/demo")
    payload = build_request_handoff_manifest(
        request_path=tmp_path / "request.yaml",
        request=request,
        run_root=tmp_path / "run",
        generated_dir=tmp_path / "run" / "generated",
        filter_dir=tmp_path / "run" / "filter",
        filtered_corpus_dir=tmp_path / "run" / "curated",
        effective_config_path=tmp_path / "run" / "generated" / "effective_config.yaml",
        effective_config_trace_path=tmp_path / "run" / "generated" / "effective_config_trace.yaml",
        filter_manifest_path=tmp_path / "run" / "filter" / "filter_manifest.ndjson",
        filter_summary_path=tmp_path / "run" / "filter" / "filter_summary.json",
        generated_datasets=2,
        generation_elapsed_seconds=12.0,
        filter_total_datasets=2,
        filter_accepted_datasets=1,
        filter_rejected_datasets=1,
        filter_elapsed_seconds=6.0,
        requested_device="cpu",
        resolved_device="cpu",
        hardware_backend="cpu",
        hardware_device_name="CPU",
        hardware_tier="cpu",
        hardware_policy="none",
    )

    validate_request_handoff_manifest(payload)

    assert payload["schema_name"] == REQUEST_HANDOFF_SCHEMA_NAME
    assert payload["schema_version"] == REQUEST_HANDOFF_SCHEMA_VERSION
    assert payload["request"]["payload"] == request.to_dict()
    assert payload["artifacts"]["filtered_corpus_dir"] == str(
        (tmp_path / "run" / "curated").resolve()
    )
    assert payload["summary"]["acceptance_rate"] == pytest.approx(0.5)
    assert payload["throughput"]["generation_stage"]["datasets_per_minute"] == pytest.approx(10.0)
    assert payload["throughput"]["filter_stage"]["datasets_per_minute"] == pytest.approx(20.0)
    assert payload["diversity_artifacts"] == {
        "summary_json_path": None,
        "summary_md_path": None,
    }


def test_write_request_handoff_manifest_writes_json_and_rejects_invalid_payload(tmp_path) -> None:
    request = _request_config(output_root="requests/demo")
    manifest_path = write_request_handoff_manifest(
        request_path=tmp_path / "request.yaml",
        request=request,
        run_root=tmp_path / "run",
        generated_dir=tmp_path / "run" / "generated",
        filter_dir=tmp_path / "run" / "filter",
        filtered_corpus_dir=tmp_path / "run" / "curated",
        effective_config_path=tmp_path / "run" / "generated" / "effective_config.yaml",
        effective_config_trace_path=tmp_path / "run" / "generated" / "effective_config_trace.yaml",
        filter_manifest_path=tmp_path / "run" / "filter" / "filter_manifest.ndjson",
        filter_summary_path=tmp_path / "run" / "filter" / "filter_summary.json",
        generated_datasets=2,
        generation_elapsed_seconds=12.0,
        filter_total_datasets=2,
        filter_accepted_datasets=2,
        filter_rejected_datasets=0,
        filter_elapsed_seconds=4.0,
        requested_device="cpu",
        resolved_device="cpu",
        hardware_backend="cpu",
        hardware_device_name="CPU",
        hardware_tier="cpu",
        hardware_policy="none",
        out_path=tmp_path / "run" / "handoff_manifest.json",
    )

    payload = json.loads(manifest_path.read_text(encoding="utf-8"))
    validate_request_handoff_manifest(payload)
    assert payload["summary"]["accepted_datasets"] == 2
    assert payload["throughput"]["filter_stage"]["datasets_per_minute"] == pytest.approx(30.0)
    assert list((tmp_path / "run").glob(".handoff_manifest.json.*.tmp")) == []

    payload["request"]["payload"]["version"] = "v2"
    with pytest.raises(
        ValueError, match=r"handoff_manifest.request.payload: version must be one of"
    ):
        validate_request_handoff_manifest(payload)


def test_write_request_handoff_manifest_does_not_overwrite_existing_file(tmp_path) -> None:
    request = _request_config(output_root="requests/demo")
    manifest_path = tmp_path / "run" / "handoff_manifest.json"
    manifest_path.parent.mkdir(parents=True, exist_ok=True)
    manifest_path.write_text('{"sentinel": true}\n', encoding="utf-8")

    with pytest.raises(RuntimeError, match="promotion target already exists"):
        _ = write_request_handoff_manifest(
            request_path=tmp_path / "request.yaml",
            request=request,
            run_root=tmp_path / "run",
            generated_dir=tmp_path / "run" / "generated",
            filter_dir=tmp_path / "run" / "filter",
            filtered_corpus_dir=tmp_path / "run" / "curated",
            effective_config_path=tmp_path / "run" / "generated" / "effective_config.yaml",
            effective_config_trace_path=tmp_path
            / "run"
            / "generated"
            / "effective_config_trace.yaml",
            filter_manifest_path=tmp_path / "run" / "filter" / "filter_manifest.ndjson",
            filter_summary_path=tmp_path / "run" / "filter" / "filter_summary.json",
            generated_datasets=2,
            generation_elapsed_seconds=12.0,
            filter_total_datasets=2,
            filter_accepted_datasets=2,
            filter_rejected_datasets=0,
            filter_elapsed_seconds=4.0,
            requested_device="cpu",
            resolved_device="cpu",
            hardware_backend="cpu",
            hardware_device_name="CPU",
            hardware_tier="cpu",
            hardware_policy="none",
            out_path=manifest_path,
        )

    assert manifest_path.read_text(encoding="utf-8") == '{"sentinel": true}\n'
    assert list(manifest_path.parent.glob(".handoff_manifest.json.*.tmp")) == []



"""Tests for math/random_matrices.py."""

import pytest
import torch
from conftest import make_generator as _make_generator

from dagzoo.math.random_matrices import sample_random_matrix


def test_output_shape() -> None:
    g = _make_generator(0)
    m = sample_random_matrix(5, 3, g, "cpu")
    assert m.shape == (5, 3)


def test_rows_unit_normalized() -> None:
    g = _make_generator(1)
    m = sample_random_matrix(4, 6, g, "cpu")
    norms = torch.linalg.norm(m, dim=1)
    torch.testing.assert_close(norms, torch.ones(4), atol=1e-4, rtol=1e-4)


def test_deterministic() -> None:
    a = sample_random_matrix(3, 4, _make_generator(42), "cpu")
    b = sample_random_matrix(3, 4, _make_generator(42), "cpu")
    torch.testing.assert_close(a, b)


@pytest.mark.parametrize("kind", ["gaussian", "weights", "singular_values", "kernel", "activation"])
def test_each_kind(kind: str) -> None:
    g = _make_generator(10)
    m = sample_random_matrix(4, 5, g, "cpu", kind=kind)
    assert m.shape == (4, 5)
    assert torch.all(torch.isfinite(m))


def test_invalid_kind_raises() -> None:
    with pytest.raises(ValueError, match="Unknown matrix kind"):
        sample_random_matrix(2, 2, _make_generator(0), "cpu", kind="bogus")



"""Request-run handoff manifest helpers for downstream corpus consumers."""

from __future__ import annotations

import json
import math
from collections.abc import Mapping
from pathlib import Path
from time import time_ns
from typing import Any, cast

from dagzoo.config import RequestFileConfig
from dagzoo.core.staged_artifacts import cleanup_path, promote_staged_path, staged_output_path
from dagzoo.math import sanitize_json

REQUEST_HANDOFF_MANIFEST_FILENAME = "handoff_manifest.json"
REQUEST_HANDOFF_SCHEMA_NAME = "dagzoo_request_handoff_manifest"
REQUEST_HANDOFF_SCHEMA_VERSION = 1


def _raise(path: str, message: str) -> None:
    raise ValueError(f"{path}: {message}")


def _require_mapping(value: object, *, path: str) -> Mapping[str, Any]:
    if not isinstance(value, Mapping):
        _raise(path, "must be a mapping")
    return cast(Mapping[str, Any], value)


def _require_non_empty_string(value: object, *, path: str) -> str:
    if not isinstance(value, str) or not value.strip():
        _raise(path, "must be a non-empty string")
    return cast(str, value)


def _require_optional_string(value: object, *, path: str) -> str | None:
    if value is None:
        return None
    return _require_non_empty_string(value, path=path)


def _require_int(value: object, *, path: str) -> int:
    if isinstance(value, bool) or not isinstance(value, int):
        _raise(path, "must be an integer")
    return int(cast(int, value))


def _require_non_negative_float(value: object, *, path: str) -> float:
    if isinstance(value, bool) or not isinstance(value, (int, float)):
        _raise(path, "must be a finite non-negative number")
    number = float(cast(int | float, value))
    if not math.isfinite(number) or number < 0.0:
        _raise(path, "must be a finite non-negative number")
    return number


def _require_optional_unit_interval(value: object, *, path: str) -> float | None:
    if value is None:
        return None
    number = _require_non_negative_float(value, path=path)
    if number > 1.0:
        _raise(path, "must be <= 1.0")
    return number


def _resolve_path_str(path: str | Path) -> str:
    return str(Path(path).resolve())


def _resolve_optional_path_str(path: str | Path | None) -> str | None:
    if path is None:
        return None
    return _resolve_path_str(path)


def _datasets_per_minute(*, datasets: int, elapsed_seconds: float) -> float:
    if elapsed_seconds <= 0.0:
        return 0.0
    return (float(datasets) / float(elapsed_seconds)) * 60.0


def build_request_handoff_manifest(
    *,
    request_path: str | Path,
    request: RequestFileConfig,
    run_root: str | Path,
    generated_dir: str | Path,
    filter_dir: str | Path,
    filtered_corpus_dir: str | Path,
    effective_config_path: str | Path,
    effective_config_trace_path: str | Path,
    filter_manifest_path: str | Path,
    filter_summary_path: str | Path,
    generated_datasets: int,
    generation_elapsed_seconds: float,
    filter_total_datasets: int,
    filter_accepted_datasets: int,
    filter_rejected_datasets: int,
    filter_elapsed_seconds: float,
    requested_device: str,
    resolved_device: str,
    hardware_backend: str,
    hardware_device_name: str,
    hardware_tier: str,
    hardware_policy: str,
    diversity_summary_json_path: str | Path | None = None,
    diversity_summary_md_path: str | Path | None = None,
) -> dict[str, Any]:
    """Build the machine-readable request handoff manifest payload."""

    total_datasets = int(filter_total_datasets)
    acceptance_rate = (
        float(filter_accepted_datasets) / float(total_datasets) if total_datasets > 0 else None
    )
    payload: dict[str, Any] = {
        "schema_name": REQUEST_HANDOFF_SCHEMA_NAME,
        "schema_version": REQUEST_HANDOFF_SCHEMA_VERSION,
        "request": {
            "path": _resolve_path_str(request_path),
            "payload": request.to_dict(),
        },
        "artifacts": {
            "run_root": _resolve_path_str(run_root),
            "generated_dir": _resolve_path_str(generated_dir),
            "filter_dir": _resolve_path_str(filter_dir),
            "filtered_corpus_dir": _resolve_path_str(filtered_corpus_dir),
            "effective_config_path": _resolve_path_str(effective_config_path),
            "effective_config_trace_path": _resolve_path_str(effective_config_trace_path),
            "filter_manifest_path": _resolve_path_str(filter_manifest_path),
            "filter_summary_path": _resolve_path_str(filter_summary_path),
        },
        "summary": {
            "generated_datasets": int(generated_datasets),
            "accepted_datasets": int(filter_accepted_datasets),
            "rejected_datasets": int(filter_rejected_datasets),
            "acceptance_rate": acceptance_rate,
        },
        "throughput": {
            "generation_stage": {
                "generated_datasets": int(generated_datasets),
                "elapsed_seconds": float(generation_elapsed_seconds),
                "datasets_per_minute": _datasets_per_minute(
                    datasets=int(generated_datasets),
                    elapsed_seconds=float(generation_elapsed_seconds),
                ),
            },
            "filter_stage": {
                "total_datasets": int(filter_total_datasets),
                "accepted_datasets": int(filter_accepted_datasets),
                "rejected_datasets": int(filter_rejected_datasets),
                "elapsed_seconds": float(filter_elapsed_seconds),
                "datasets_per_minute": _datasets_per_minute(
                    datasets=int(filter_total_datasets),
                    elapsed_seconds=float(filter_elapsed_seconds),
                ),
            },
        },
        "hardware": {
            "requested_device": str(requested_device),
            "resolved_device": str(resolved_device),
            "backend": str(hardware_backend),
            "device_name": str(hardware_device_name),
            "tier": str(hardware_tier),
            "hardware_policy": str(hardware_policy),
        },
        "diversity_artifacts": {
            "summary_json_path": _resolve_optional_path_str(diversity_summary_json_path),
            "summary_md_path": _resolve_optional_path_str(diversity_summary_md_path),
        },
    }
    validate_request_handoff_manifest(payload)
    return payload


def validate_request_handoff_manifest(payload: Mapping[str, Any]) -> None:
    """Validate the request handoff manifest wire shape."""

    root = _require_mapping(payload, path="handoff_manifest")
    schema_name = _require_non_empty_string(
        root.get("schema_name"),
        path="handoff_manifest.schema_name",
    )
    if schema_name != REQUEST_HANDOFF_SCHEMA_NAME:
        _raise(
            "handoff_manifest.schema_name",
            f"must equal {REQUEST_HANDOFF_SCHEMA_NAME!r}",
        )
    schema_version = _require_int(
        root.get("schema_version"),
        path="handoff_manifest.schema_version",
    )
    if schema_version != REQUEST_HANDOFF_SCHEMA_VERSION:
        _raise(
            "handoff_manifest.schema_version",
            f"must equal {REQUEST_HANDOFF_SCHEMA_VERSION}",
        )

    request = _require_mapping(root.get("request"), path="handoff_manifest.request")
    _require_non_empty_string(request.get("path"), path="handoff_manifest.request.path")
    request_payload = _require_mapping(
        request.get("payload"),
        path="handoff_manifest.request.payload",
    )
    try:
        RequestFileConfig.from_dict(dict(request_payload))
    except (TypeError, ValueError) as exc:
        _raise("handoff_manifest.request.payload", str(exc))

    artifacts = _require_mapping(root.get("artifacts"), path="handoff_manifest.artifacts")
    for key in (
        "run_root",
        "generated_dir",
        "filter_dir",
        "filtered_corpus_dir",
        "effective_config_path",
        "effective_config_trace_path",
        "filter_manifest_path",
        "filter_summary_path",
    ):
        _require_non_empty_string(
            artifacts.get(key),
            path=f"handoff_manifest.artifacts.{key}",
        )

    summary = _require_mapping(root.get("summary"), path="handoff_manifest.summary")
    _require_int(
        summary.get("generated_datasets"), path="handoff_manifest.summary.generated_datasets"
    )
    _require_int(
        summary.get("accepted_datasets"), path="handoff_manifest.summary.accepted_datasets"
    )
    _require_int(
        summary.get("rejected_datasets"), path="handoff_manifest.summary.rejected_datasets"
    )
    _require_optional_unit_interval(
        summary.get("acceptance_rate"),
        path="handoff_manifest.summary.acceptance_rate",
    )

    throughput = _require_mapping(root.get("throughput"), path="handoff_manifest.throughput")
    generation_stage = _require_mapping(
        throughput.get("generation_stage"),
        path="handoff_manifest.throughput.generation_stage",
    )
    _require_int(
        generation_stage.get("generated_datasets"),
        path="handoff_manifest.throughput.generation_stage.generated_datasets",
    )
    _require_non_negative_float(
        generation_stage.get("elapsed_seconds"),
        path="handoff_manifest.throughput.generation_stage.elapsed_seconds",
    )
    _require_non_negative_float(
        generation_stage.get("datasets_per_minute"),
        path="handoff_manifest.throughput.generation_stage.datasets_per_minute",
    )

    filter_stage = _require_mapping(
        throughput.get("filter_stage"),
        path="handoff_manifest.throughput.filter_stage",
    )
    _require_int(
        filter_stage.get("total_datasets"),
        path="handoff_manifest.throughput.filter_stage.total_datasets",
    )
    _require_int(
        filter_stage.get("accepted_datasets"),
        path="handoff_manifest.throughput.filter_stage.accepted_datasets",
    )
    _require_int(
        filter_stage.get("rejected_datasets"),
        path="handoff_manifest.throughput.filter_stage.rejected_datasets",
    )
    _require_non_negative_float(
        filter_stage.get("elapsed_seconds"),
        path="handoff_manifest.throughput.filter_stage.elapsed_seconds",
    )
    _require_non_negative_float(
        filter_stage.get("datasets_per_minute"),
        path="handoff_manifest.throughput.filter_stage.datasets_per_minute",
    )

    hardware = _require_mapping(root.get("hardware"), path="handoff_manifest.hardware")
    for key in (
        "requested_device",
        "resolved_device",
        "backend",
        "device_name",
        "tier",
        "hardware_policy",
    ):
        _require_non_empty_string(
            hardware.get(key),
            path=f"handoff_manifest.hardware.{key}",
        )

    diversity_artifacts = _require_mapping(
        root.get("diversity_artifacts"),
        path="handoff_manifest.diversity_artifacts",
    )
    _require_optional_string(
        diversity_artifacts.get("summary_json_path"),
        path="handoff_manifest.diversity_artifacts.summary_json_path",
    )
    _require_optional_string(
        diversity_artifacts.get("summary_md_path"),
        path="handoff_manifest.diversity_artifacts.summary_md_path",
    )


def write_request_handoff_manifest(
    *,
    request_path: str | Path,
    request: RequestFileConfig,
    run_root: str | Path,
    generated_dir: str | Path,
    filter_dir: str | Path,
    filtered_corpus_dir: str | Path,
    effective_config_path: str | Path,
    effective_config_trace_path: str | Path,
    filter_manifest_path: str | Path,
    filter_summary_path: str | Path,
    generated_datasets: int,
    generation_elapsed_seconds: float,
    filter_total_datasets: int,
    filter_accepted_datasets: int,
    filter_rejected_datasets: int,
    filter_elapsed_seconds: float,
    requested_device: str,
    resolved_device: str,
    hardware_backend: str,
    hardware_device_name: str,
    hardware_tier: str,
    hardware_policy: str,
    diversity_summary_json_path: str | Path | None = None,
    diversity_summary_md_path: str | Path | None = None,
    out_path: str | Path | None = None,
) -> Path:
    """Write the request handoff manifest to disk and return its path."""

    manifest_path = (
        Path(out_path)
        if out_path is not None
        else Path(run_root) / REQUEST_HANDOFF_MANIFEST_FILENAME
    )
    payload = build_request_handoff_manifest(
        request_path=request_path,
        request=request,
        run_root=run_root,
        generated_dir=generated_dir,
        filter_dir=filter_dir,
        filtered_corpus_dir=filtered_corpus_dir,
        effective_config_path=effective_config_path,
        effective_config_trace_path=effective_config_trace_path,
        filter_manifest_path=filter_manifest_path,
        filter_summary_path=filter_summary_path,
        generated_datasets=generated_datasets,
        generation_elapsed_seconds=generation_elapsed_seconds,
        filter_total_datasets=filter_total_datasets,
        filter_accepted_datasets=filter_accepted_datasets,
        filter_rejected_datasets=filter_rejected_datasets,
        filter_elapsed_seconds=filter_elapsed_seconds,
        requested_device=requested_device,
        resolved_device=resolved_device,
        hardware_backend=hardware_backend,
        hardware_device_name=hardware_device_name,
        hardware_tier=hardware_tier,
        hardware_policy=hardware_policy,
        diversity_summary_json_path=diversity_summary_json_path,
        diversity_summary_md_path=diversity_summary_md_path,
    )
    manifest_path.parent.mkdir(parents=True, exist_ok=True)
    staged_manifest_path = staged_output_path(
        parent_dir=manifest_path.parent,
        final_name=manifest_path.name,
        staging_token=str(time_ns()),
    )
    try:
        staged_manifest_path.write_text(
            json.dumps(sanitize_json(payload), indent=2, sort_keys=True, allow_nan=False) + "\n",
            encoding="utf-8",
        )
        promote_staged_path(staged_path=staged_manifest_path, final_path=manifest_path)
    finally:
        cleanup_path(staged_manifest_path)
    return manifest_path


__all__ = [
    "REQUEST_HANDOFF_MANIFEST_FILENAME",
    "REQUEST_HANDOFF_SCHEMA_NAME",
    "REQUEST_HANDOFF_SCHEMA_VERSION",
    "build_request_handoff_manifest",
    "validate_request_handoff_manifest",
    "write_request_handoff_manifest",
]



"""Public request-file contract for downstream handoff workflows."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, cast

import yaml

from dagzoo.config.constants import (
    MISSINGNESS_MECHANISM_MAR,
    MISSINGNESS_MECHANISM_MCAR,
    MISSINGNESS_MECHANISM_MNAR,
    MISSINGNESS_MECHANISM_NONE,
)
from dagzoo.config.rows import DatasetRowsSpec, normalize_dataset_rows
from dagzoo.rng import SEED32_MAX, SEED32_MIN

REQUEST_FILE_VERSION_V1 = "v1"
REQUEST_PROFILE_DEFAULT = "default"
REQUEST_PROFILE_SMOKE = "smoke"
REQUEST_TASK_CLASSIFICATION = "classification"
REQUEST_TASK_REGRESSION = "regression"

_REQUEST_ALLOWED_KEYS = {
    "dataset_count",
    "missingness_profile",
    "output_root",
    "profile",
    "rows",
    "seed",
    "task",
    "version",
}
_REQUEST_FORBIDDEN_KEYS = {
    "benchmark",
    "dataset",
    "diagnostics",
    "filter",
    "graph",
    "mechanism",
    "n_test",
    "n_train",
    "noise",
    "output",
    "runtime",
    "shift",
}
_REQUEST_PROFILES = {REQUEST_PROFILE_DEFAULT, REQUEST_PROFILE_SMOKE}
_REQUEST_TASKS = {REQUEST_TASK_CLASSIFICATION, REQUEST_TASK_REGRESSION}
_REQUEST_MISSINGNESS_PROFILES = {
    MISSINGNESS_MECHANISM_NONE,
    MISSINGNESS_MECHANISM_MCAR,
    MISSINGNESS_MECHANISM_MAR,
    MISSINGNESS_MECHANISM_MNAR,
}
_REQUEST_ROWS_PUBLIC_SHAPE_ERROR = (
    "rows must use one of the public request-file encodings: "
    "fixed integer, range string start..stop, or CSV choice string."
)


def _validate_request_field_names(data: dict[object, Any]) -> None:
    """Reject non-string top-level request fields before unknown-key handling."""

    for key in data:
        if not isinstance(key, str):
            raise ValueError(f"Request-file field names must be strings (got {key!r}).")


def _validate_non_empty_string(*, field_name: str, value: object) -> str:
    """Validate one required non-empty string field."""

    if not isinstance(value, str):
        raise ValueError(f"{field_name} must be a non-empty string.")
    normalized = value.strip()
    if not normalized:
        raise ValueError(f"{field_name} must be a non-empty string.")
    return normalized


def _validate_positive_int(*, field_name: str, value: object) -> int:
    """Validate one strictly positive integer field."""

    if isinstance(value, bool) or not isinstance(value, int):
        raise ValueError(f"{field_name} must be a positive integer.")
    if value <= 0:
        raise ValueError(f"{field_name} must be a positive integer.")
    return int(value)


def _validate_seed(*, field_name: str, value: object) -> int:
    """Validate a 32-bit request seed."""

    if isinstance(value, bool) or not isinstance(value, int):
        raise ValueError(f"{field_name} must be an integer in [{SEED32_MIN}, {SEED32_MAX}].")
    if value < SEED32_MIN or value > SEED32_MAX:
        raise ValueError(f"{field_name} must be an integer in [{SEED32_MIN}, {SEED32_MAX}].")
    return int(value)


def _validate_choice(*, field_name: str, value: object, choices: set[str]) -> str:
    """Validate one case-insensitive enum-like string field."""

    normalized = _validate_non_empty_string(field_name=field_name, value=value).lower()
    if normalized not in choices:
        expected = ", ".join(sorted(choices))
        raise ValueError(f"{field_name} must be one of: {expected} (got {value!r}).")
    return normalized


def _require_request_field(data: dict[str, Any], field_name: str) -> Any:
    """Return one required request field or raise a consistent error."""

    if field_name not in data:
        raise ValueError(f"Missing required request-file field: {field_name!r}.")
    return data[field_name]


def _normalize_public_request_rows(value: int | str) -> DatasetRowsSpec:
    """Normalize one public rows value and rewrite internal field names."""

    try:
        rows = normalize_dataset_rows(value)
    except ValueError as exc:
        raise ValueError(str(exc).replace("dataset.rows", "rows")) from None
    if rows is None:
        raise ValueError(_REQUEST_ROWS_PUBLIC_SHAPE_ERROR)
    return rows


def _validate_public_request_rows(value: object) -> DatasetRowsSpec:
    """Validate rows using only the documented public request-file encodings."""

    if isinstance(value, bool):
        raise ValueError(_REQUEST_ROWS_PUBLIC_SHAPE_ERROR)
    if isinstance(value, int):
        return _normalize_public_request_rows(value)
    if isinstance(value, str):
        normalized = value.strip()
        if ".." not in normalized and "," not in normalized:
            raise ValueError(_REQUEST_ROWS_PUBLIC_SHAPE_ERROR)
        return _normalize_public_request_rows(normalized)
    raise ValueError(_REQUEST_ROWS_PUBLIC_SHAPE_ERROR)


def _serialize_public_request_rows(rows: DatasetRowsSpec) -> int | str:
    """Serialize a normalized rows spec to the public request-file wire shape."""

    if rows.mode == "fixed":
        if rows.value is None:
            raise ValueError("rows fixed mode requires a value.")
        return int(rows.value)
    if rows.mode == "range":
        if rows.start is None or rows.stop is None:
            raise ValueError("rows range mode requires start/stop.")
        return f"{int(rows.start)}..{int(rows.stop)}"
    if rows.mode == "choices":
        if not rows.choices:
            raise ValueError("rows choices mode requires at least one choice.")
        return ",".join(str(int(choice)) for choice in rows.choices)
    raise ValueError(f"Unsupported request rows mode {rows.mode!r}.")


@dataclass(slots=True)
class RequestFileConfig:
    """Small public request-file schema for downstream corpus requests."""

    version: str
    task: str
    dataset_count: int
    rows: object
    profile: str
    output_root: str
    missingness_profile: str = MISSINGNESS_MECHANISM_NONE
    seed: int | None = None

    def __post_init__(self) -> None:
        self.version = _validate_choice(
            field_name="version",
            value=self.version,
            choices={REQUEST_FILE_VERSION_V1},
        )
        self.task = _validate_choice(
            field_name="task",
            value=self.task,
            choices=_REQUEST_TASKS,
        )
        self.dataset_count = _validate_positive_int(
            field_name="dataset_count",
            value=self.dataset_count,
        )
        self.rows = _validate_public_request_rows(self.rows)
        self.profile = _validate_choice(
            field_name="profile",
            value=self.profile,
            choices=_REQUEST_PROFILES,
        )
        self.output_root = _validate_non_empty_string(
            field_name="output_root",
            value=self.output_root,
        )
        self.missingness_profile = _validate_choice(
            field_name="missingness_profile",
            value=self.missingness_profile,
            choices=_REQUEST_MISSINGNESS_PROFILES,
        )
        if self.seed is not None:
            self.seed = _validate_seed(field_name="seed", value=self.seed)

    @classmethod
    def from_dict(cls, data: dict[str, Any] | None) -> RequestFileConfig:
        """Construct a request-file config from a mapping payload."""

        data = data or {}
        _validate_request_field_names(cast(dict[object, Any], data))
        unknown_keys = sorted(set(data) - _REQUEST_ALLOWED_KEYS)
        if unknown_keys:
            first_key = unknown_keys[0]
            if first_key in _REQUEST_FORBIDDEN_KEYS:
                raise ValueError(
                    f"{first_key} is not part of the public request-file contract. "
                    "Use the minimal request fields instead of raw internal config sections."
                )
            raise ValueError(f"Unknown request-file field: {first_key!r}.")
        forbidden_present = sorted(set(data).intersection(_REQUEST_FORBIDDEN_KEYS))
        if forbidden_present:
            first_key = forbidden_present[0]
            raise ValueError(
                f"{first_key} is not part of the public request-file contract. "
                "Use the minimal request fields instead of raw internal config sections."
            )

        return cls(
            version=cast(str, _require_request_field(data, "version")),
            task=cast(str, _require_request_field(data, "task")),
            dataset_count=cast(int, _require_request_field(data, "dataset_count")),
            rows=_require_request_field(data, "rows"),
            profile=cast(str, _require_request_field(data, "profile")),
            output_root=cast(str, _require_request_field(data, "output_root")),
            missingness_profile=data.get("missingness_profile", MISSINGNESS_MECHANISM_NONE),
            seed=data.get("seed"),
        )

    @classmethod
    def from_yaml(cls, path: str | Path) -> RequestFileConfig:
        """Load a request-file config from a YAML file path."""

        resolved_path = Path(path)
        with resolved_path.open("r", encoding="utf-8") as handle:
            loaded = yaml.safe_load(handle) or {}
        if not isinstance(loaded, dict):
            raise ValueError(f"Request file at {resolved_path} must be a mapping at the top level.")
        return cls.from_dict(loaded)

    def to_dict(self) -> dict[str, Any]:
        """Serialize request config to a plain nested dictionary."""

        rows = self.rows
        if not isinstance(rows, DatasetRowsSpec):
            raise ValueError("rows must be normalized before serialization.")

        data: dict[str, Any] = {
            "version": self.version,
            "task": self.task,
            "dataset_count": self.dataset_count,
            "rows": _serialize_public_request_rows(rows),
            "profile": self.profile,
            "output_root": self.output_root,
            "missingness_profile": self.missingness_profile,
        }
        if self.seed is not None:
            data["seed"] = self.seed
        return data



from pathlib import Path

import pytest

from dagzoo.cli import main


def test_generate_cli_prints_effective_config_and_resolution_trace(
    tmp_path, capsys: pytest.CaptureFixture[str], monkeypatch: pytest.MonkeyPatch
) -> None:
    def _stub_generate_batch_iter(
        _config,
        *,
        num_datasets: int,
        seed: int | None = None,
        device: str | None = None,
    ):
        _ = seed
        _ = device
        for _ in range(num_datasets):
            yield object()

    monkeypatch.setattr("dagzoo.cli.generate_batch_iter", _stub_generate_batch_iter)

    code = main(
        [
            "generate",
            "--config",
            "configs/default.yaml",
            "--out",
            str(tmp_path / "generate"),
            "--num-datasets",
            "1",
            "--device",
            "cpu",
            "--hardware-policy",
            "none",
            "--no-dataset-write",
            "--print-effective-config",
            "--print-resolution-trace",
        ]
    )

    assert code == 0
    captured = capsys.readouterr()
    assert "Effective config:" in captured.out
    assert "Resolution trace:" in captured.out


def test_filter_cli_prints_curated_output_summary(
    tmp_path, capsys: pytest.CaptureFixture[str], monkeypatch: pytest.MonkeyPatch
) -> None:
    class _Result:
        manifest_path = Path("manifest.ndjson")
        summary_path = Path("summary.json")
        total_datasets = 2
        accepted_datasets = 1
        rejected_datasets = 1
        datasets_per_minute = 42.0
        curated_out_dir = Path("curated")
        curated_accepted_datasets = 1

    monkeypatch.setattr("dagzoo.cli.run_deferred_filter", lambda **kwargs: _Result())

    code = main(
        [
            "filter",
            "--in",
            "input_shards",
            "--out",
            str(tmp_path / "filter_out"),
            "--curated-out",
            str(tmp_path / "curated_out"),
        ]
    )

    assert code == 0
    captured = capsys.readouterr()
    assert "Wrote curated accepted-only shards:" in captured.out


def test_request_cli_prints_request_execution_summary(
    tmp_path, capsys: pytest.CaptureFixture[str], monkeypatch: pytest.MonkeyPatch
) -> None:
    class _FilterResult:
        manifest_path = Path("filter_manifest.ndjson")
        summary_path = Path("filter_summary.json")
        total_datasets = 2
        accepted_datasets = 1
        rejected_datasets = 1
        datasets_per_minute = 42.0
        curated_out_dir = Path("curated")
        curated_accepted_datasets = 1

    class _Result:
        effective_config_path = Path("effective_config.yaml")
        effective_config_trace_path = Path("effective_config_trace.yaml")
        handoff_manifest_path = Path("handoff_manifest.json")
        generated_dir = Path("generated")
        generated_datasets = 2
        filter_result = _FilterResult()

    monkeypatch.setattr("dagzoo.cli.run_request_execution", lambda **kwargs: _Result())

    request_path = tmp_path / "request.yaml"
    request_path.write_text(
        "version: v1\n"
        "task: classification\n"
        "dataset_count: 2\n"
        "rows: 1024\n"
        "profile: default\n"
        "output_root: requests/out\n",
        encoding="utf-8",
    )

    code = main(
        [
            "request",
            "--request",
            str(request_path),
        ]
    )

    assert code == 0
    captured = capsys.readouterr()
    assert "Wrote effective config:" in captured.out
    assert "Wrote handoff manifest:" in captured.out
    assert "Wrote filter manifest:" in captured.out
    assert "Deferred filter summary:" in captured.out
    assert "Wrote curated accepted-only shards:" in captured.out


def test_benchmark_cli_prints_configs_and_writes_baseline(
    tmp_path, capsys: pytest.CaptureFixture[str], monkeypatch: pytest.MonkeyPatch
) -> None:
    summary = {
        "preset_results": [
            {
                "preset_key": "custom",
                "effective_config": {"runtime": {"device": "cpu"}},
                "effective_config_trace": [
                    {
                        "path": "runtime.device",
                        "source": "config",
                        "old_value": "auto",
                        "new_value": "cpu",
                    }
                ],
                "datasets_per_minute": 1.0,
                "latency_p95_ms": 1.0,
            }
        ],
        "regression": {"status": "pass", "issues": [], "hard_fail": False},
    }

    monkeypatch.setattr("dagzoo.cli.run_benchmark_suite", lambda *args, **kwargs: summary)
    monkeypatch.setattr("dagzoo.cli.write_suite_json", lambda _summary, path: Path(path))
    monkeypatch.setattr(
        "dagzoo.cli._print_preset_result_line",
        lambda _result: None,
    )
    monkeypatch.setattr("dagzoo.cli.build_baseline_payload", lambda _summary: {"schema_version": 1})
    monkeypatch.setattr("dagzoo.cli.write_baseline", lambda _payload, path: Path(path))

    code = main(
        [
            "benchmark",
            "--config",
            "configs/default.yaml",
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--json-out",
            str(tmp_path / "summary.json"),
            "--save-baseline",
            str(tmp_path / "baseline.json"),
            "--print-effective-config",
            "--print-resolution-trace",
            "--no-memory",
        ]
    )

    assert code == 0
    captured = capsys.readouterr()
    assert "Effective config [custom]:" in captured.out
    assert "Resolution trace [custom]:" in captured.out
    assert "Wrote benchmark baseline:" in captured.out


def test_diversity_audit_cli_writes_summary_and_status(
    tmp_path, capsys: pytest.CaptureFixture[str], monkeypatch: pytest.MonkeyPatch
) -> None:
    report = {
        "summary": {"overall_status": "warn", "num_variants": 1},
    }

    monkeypatch.setattr("dagzoo.cli.run_effective_diversity_audit", lambda **kwargs: report)
    monkeypatch.setattr(
        "dagzoo.cli.write_effective_diversity_artifacts",
        lambda _report, out_dir: {"summary_json": Path(out_dir) / "summary.json"},
    )

    code = main(
        [
            "diversity-audit",
            "--baseline-config",
            "configs/default.yaml",
            "--variant-config",
            "configs/preset_shift_benchmark_smoke.yaml",
            "--out-dir",
            str(tmp_path / "diversity"),
        ]
    )

    assert code == 0
    captured = capsys.readouterr()
    assert "Wrote diversity artifact [summary_json]:" in captured.out
    assert "Diversity audit status=warn variants=1" in captured.out


def test_diversity_audit_cli_fail_on_regression_treats_insufficient_metrics_as_error(
    tmp_path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(
        "dagzoo.cli.run_effective_diversity_audit",
        lambda **_kwargs: {
            "summary": {"overall_status": "insufficient_metrics", "num_variants": 1}
        },
    )
    monkeypatch.setattr(
        "dagzoo.cli.write_effective_diversity_artifacts",
        lambda _report, out_dir: {"summary_json": Path(out_dir) / "summary.json"},
    )

    code = main(
        [
            "diversity-audit",
            "--baseline-config",
            "configs/default.yaml",
            "--variant-config",
            "configs/preset_shift_benchmark_smoke.yaml",
            "--fail-on-regression",
            "--out-dir",
            str(tmp_path / "diversity"),
        ]
    )

    assert code == 1


def test_filter_calibration_cli_writes_summary_and_status(
    tmp_path, capsys: pytest.CaptureFixture[str], monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(
        "dagzoo.cli.run_filter_calibration",
        lambda **_kwargs: {
            "summary": {
                "overall_status": "warn",
                "best_overall_threshold_requested": 0.804,
                "best_overall_diversity_status": "warn",
                "best_passing_threshold_requested": 0.801,
                "num_candidates": 5,
            }
        },
    )
    monkeypatch.setattr(
        "dagzoo.cli.write_filter_calibration_artifacts",
        lambda _report, out_dir: {"summary_json": Path(out_dir) / "summary.json"},
    )

    code = main(
        [
            "filter-calibration",
            "--config",
            "configs/preset_filter_benchmark_smoke.yaml",
            "--out-dir",
            str(tmp_path / "filter_calibration"),
        ]
    )

    assert code == 0
    captured = capsys.readouterr()
    assert "Wrote filter calibration artifact [summary_json]:" in captured.out
    assert "Filter calibration status=warn best_overall=0.804 best_passing=0.801 candidates=5" in (
        captured.out
    )


def test_filter_calibration_cli_fail_on_regression_uses_best_overall_status(
    tmp_path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(
        "dagzoo.cli.run_filter_calibration",
        lambda **_kwargs: {
            "summary": {
                "overall_status": "warn",
                "best_overall_threshold_requested": 0.8,
                "best_overall_diversity_status": "fail",
                "best_passing_threshold_requested": None,
                "num_candidates": 5,
            }
        },
    )
    monkeypatch.setattr(
        "dagzoo.cli.write_filter_calibration_artifacts",
        lambda _report, out_dir: {"summary_json": Path(out_dir) / "summary.json"},
    )

    code = main(
        [
            "filter-calibration",
            "--config",
            "configs/preset_filter_benchmark_smoke.yaml",
            "--fail-on-regression",
            "--out-dir",
            str(tmp_path / "filter_calibration"),
        ]
    )

    assert code == 1


def test_hardware_cli_prints_detected_hardware(
    capsys: pytest.CaptureFixture[str], monkeypatch: pytest.MonkeyPatch
) -> None:
    class _Hardware:
        backend = "cpu"
        device_name = "cpu"
        tier = "cpu"
        total_memory_gb = 8.0
        peak_flops = 1.23e12

    monkeypatch.setattr("dagzoo.cli.detect_hardware", lambda _device: _Hardware())

    code = main(["hardware", "--device", "cpu"])

    assert code == 0
    captured = capsys.readouterr()
    assert "backend=cpu" in captured.out
    assert "tier=cpu" in captured.out



from __future__ import annotations

RELEASE_RISK_EXACT_PATHS = frozenset(
    {
        "src/dagzoo/core/config_resolution.py",
        "src/dagzoo/core/dataset.py",
        "src/dagzoo/core/fixed_layout/runtime.py",
        "src/dagzoo/core/metadata.py",
    }
)
RELEASE_RISK_PREFIXES = (
    "src/dagzoo/cli/",
    "src/dagzoo/config/",
    "src/dagzoo/io/",
    "configs/",
)

SUGGESTED_PYTEST_TARGETS = (
    (
        "src/dagzoo/cli/",
        (
            "tests/test_cli_validation.py",
            "tests/test_cli_outputs.py",
            "tests/test_benchmark_cli.py",
            "tests/test_request_execution.py",
        ),
    ),
    (
        "src/dagzoo/bench/",
        (
            "tests/test_benchmark_suite.py",
            "tests/test_benchmark_cli.py",
            "tests/test_benchmark_stage_metrics.py",
            "tests/test_benchmark_throughput.py",
        ),
    ),
    (
        "src/dagzoo/config/",
        (
            "tests/test_config.py",
            "tests/test_config_resolution.py",
            "tests/test_request_config.py",
            "tests/test_request_execution.py",
        ),
    ),
    (
        "src/dagzoo/core/config_resolution.py",
        (
            "tests/test_config.py",
            "tests/test_config_resolution.py",
            "tests/test_request_config.py",
            "tests/test_request_execution.py",
        ),
    ),
    (
        "configs/",
        (
            "tests/test_config.py",
            "tests/test_config_resolution.py",
            "tests/test_request_execution.py",
            "tests/test_benchmark_cli.py",
        ),
    ),
    ("scripts/devlib/", ("tests/test_dev_tooling.py",)),
    (".pre-commit-config.yaml", ("tests/test_dev_tooling.py",)),
    (
        "scripts/docs/",
        (
            "tests/test_docs_scripts.py",
            "tests/test_dev_tooling.py",
        ),
    ),
    ("README.md", ("tests/test_docs_scripts.py",)),
    ("docs/", ("tests/test_docs_scripts.py",)),
    ("site/", ("tests/test_docs_scripts.py",)),
)


def is_release_risk_path(path: str) -> bool:
    return path in RELEASE_RISK_EXACT_PATHS or path.startswith(RELEASE_RISK_PREFIXES)


def suggested_pytest_targets(changed_files: tuple[str, ...]) -> tuple[str, ...]:
    seen: set[str] = set()
    suggested: list[str] = []
    for changed_file in changed_files:
        for path_prefix, targets in SUGGESTED_PYTEST_TARGETS:
            if changed_file == path_prefix or changed_file.startswith(path_prefix):
                for target in targets:
                    if target in seen:
                        continue
                    seen.add(target)
                    suggested.append(target)
    return tuple(suggested)



"""Request command handler."""

from __future__ import annotations

import argparse

import yaml

from ..common import get_cli_public_api, raise_usage_error


def run_request_command(args: argparse.Namespace) -> int:
    """Execute the ``request`` command."""

    cli_api = get_cli_public_api()
    try:
        result = cli_api.run_request_execution(
            request_path=args.request,
            device_override=args.device,
            hardware_policy=str(args.hardware_policy),
            n_jobs_override=args.n_jobs,
            print_effective_config_flag=bool(args.print_effective_config),
            print_resolution_trace_flag=bool(args.print_resolution_trace),
        )
    except yaml.YAMLError as exc:
        raise_usage_error(f"Failed to parse request file {args.request}: {exc}")
    except (FileNotFoundError, TypeError, ValueError, RuntimeError) as exc:
        raise_usage_error(str(exc))

    print(f"Wrote effective config: {result.effective_config_path}")
    print(f"Wrote effective config trace: {result.effective_config_trace_path}")
    print(f"Wrote handoff manifest: {result.handoff_manifest_path}")
    print(f"Wrote {result.generated_datasets} datasets to: {result.generated_dir}")
    print(f"Wrote filter manifest: {result.filter_result.manifest_path}")
    print(f"Wrote filter summary: {result.filter_result.summary_path}")
    print(
        "Deferred filter summary: "
        f"total={result.filter_result.total_datasets} "
        f"accepted={result.filter_result.accepted_datasets} "
        f"rejected={result.filter_result.rejected_datasets} "
        f"dpm={result.filter_result.datasets_per_minute:.2f}"
    )
    if result.filter_result.curated_out_dir is not None:
        print(
            f"Wrote curated accepted-only shards: {result.filter_result.curated_out_dir} "
            f"(datasets={result.filter_result.curated_accepted_datasets})"
        )
    return 0



"""Shared config-resolution helpers for generate and benchmark command paths."""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Any

from dagzoo.config import (
    DATASET_ROWS_MIN_TOTAL,
    MISSINGNESS_MECHANISM_MAR,
    MISSINGNESS_MECHANISM_MCAR,
    MISSINGNESS_MECHANISM_MNAR,
    MISSINGNESS_MECHANISM_NONE,
    REQUEST_PROFILE_SMOKE,
    DatasetRowsSpec,
    GeneratorConfig,
    RequestFileConfig,
    clone_generator_config,
    load_packaged_generator_config,
    normalize_dataset_rows,
)
from dagzoo.hardware import HardwareInfo, detect_hardware
from dagzoo.hardware_policy import (
    apply_hardware_policy,
    resolve_cuda_fixed_layout_target_cells_limits,
)

_MISSING_VALUE = "<missing>"
_DEFAULT_CUDA_FIXED_LAYOUT_TARGET_SOURCE = "hardware.default_cuda_fixed_layout_target_cells"
_REQUEST_BASE_CONFIG_RESOURCE = "default.yaml"
_REQUEST_MISSINGNESS_PRESET_RESOURCES = {
    MISSINGNESS_MECHANISM_MCAR: "preset_missingness_mcar.yaml",
    MISSINGNESS_MECHANISM_MAR: "preset_missingness_mar.yaml",
    MISSINGNESS_MECHANISM_MNAR: "preset_missingness_mnar.yaml",
}
_REQUEST_MISSINGNESS_DATASET_PATHS = (
    "dataset.missing_rate",
    "dataset.missing_mechanism",
    "dataset.missing_mar_observed_fraction",
    "dataset.missing_mar_logit_scale",
    "dataset.missing_mnar_logit_scale",
)


@dataclass(slots=True, frozen=True)
class ResolutionEvent:
    """One field-level config override event."""

    path: str
    source: str
    old_value: Any
    new_value: Any


@dataclass(slots=True, frozen=True)
class BenchmarkSmokeCaps:
    """Hard caps applied to benchmark preset configs in smoke mode."""

    n_train: int
    n_test: int
    n_features: int
    n_nodes: int


@dataclass(slots=True, frozen=True)
class RequestSmokeProfile:
    """Hard caps applied to request configs for the public smoke profile."""

    n_train: int
    n_test: int
    n_features_min: int
    n_features_max: int
    n_nodes_min: int
    n_nodes_max: int


@dataclass(slots=True)
class ResolvedGenerateConfig:
    """Fully resolved generate config with hardware info and override trace."""

    config: GeneratorConfig
    hardware: HardwareInfo
    requested_device: str
    trace_events: list[ResolutionEvent]


@dataclass(slots=True)
class ResolvedRequestConfig:
    """Fully resolved request config with hardware info and override trace."""

    request: RequestFileConfig
    config: GeneratorConfig
    hardware: HardwareInfo
    requested_device: str
    trace_events: list[ResolutionEvent]


@dataclass(slots=True)
class ResolvedBenchmarkPresetConfig:
    """Fully resolved benchmark preset config with hardware info and override trace."""

    preset_key: str
    config: GeneratorConfig
    hardware: HardwareInfo
    requested_device: str
    trace_events: list[ResolutionEvent]


def _clone_config(config: GeneratorConfig) -> GeneratorConfig:
    """Clone nested config state so resolution does not mutate caller-owned objects."""

    return clone_generator_config(config, revalidate=True)


def _append_event(
    *,
    events: list[ResolutionEvent],
    path: str,
    source: str,
    old_value: Any,
    new_value: Any,
) -> None:
    """Append one trace event when values differ."""

    if old_value == new_value:
        return
    events.append(
        ResolutionEvent(
            path=path,
            source=source,
            old_value=old_value,
            new_value=new_value,
        )
    )


def _set_config_path(
    config: GeneratorConfig,
    *,
    path: str,
    value: Any,
    source: str,
    events: list[ResolutionEvent],
) -> None:
    """Set a dotted dataclass path and emit one trace event when changed."""

    parts = path.split(".")
    target: Any = config
    for part in parts[:-1]:
        target = getattr(target, part)
    field_name = parts[-1]
    old_value = getattr(target, field_name)
    if old_value == value:
        return
    setattr(target, field_name, value)
    _append_event(
        events=events,
        path=path,
        source=source,
        old_value=old_value,
        new_value=value,
    )


def _append_diff_events(
    before: Any,
    after: Any,
    *,
    source: str,
    events: list[ResolutionEvent],
    path: str = "",
) -> None:
    """Append trace events by diffing serialized config payloads."""

    if isinstance(before, dict) and isinstance(after, dict):
        keys = sorted(set(before) | set(after))
        for key in keys:
            child_path = key if not path else f"{path}.{key}"
            old_value = before.get(key, _MISSING_VALUE)
            new_value = after.get(key, _MISSING_VALUE)
            _append_diff_events(
                old_value,
                new_value,
                source=source,
                events=events,
                path=child_path,
            )
        return
    _append_event(
        events=events,
        path=path,
        source=source,
        old_value=before,
        new_value=after,
    )


def _normalize_requested_device(device: str | None, fallback: str | None) -> str:
    """Normalize device text into the same lower-cased runtime shape used by CLI paths."""

    if device is not None:
        candidate = str(device).strip().lower()
        return candidate or "auto"
    if fallback is None:
        return "auto"
    candidate = str(fallback).strip().lower()
    return candidate or "auto"


def _apply_missingness_overrides(
    config: GeneratorConfig,
    *,
    missing_rate: float | None,
    missing_mechanism: str | None,
    missing_mar_observed_fraction: float | None,
    missing_mar_logit_scale: float | None,
    missing_mnar_logit_scale: float | None,
    events: list[ResolutionEvent],
) -> None:
    """Apply generate missingness overrides."""

    overrides = (
        ("dataset.missing_rate", missing_rate),
        ("dataset.missing_mechanism", missing_mechanism),
        ("dataset.missing_mar_observed_fraction", missing_mar_observed_fraction),
        ("dataset.missing_mar_logit_scale", missing_mar_logit_scale),
        ("dataset.missing_mnar_logit_scale", missing_mnar_logit_scale),
    )
    has_override = any(value is not None for _, value in overrides)
    if not has_override:
        return

    for path, value in overrides:
        if value is None:
            continue
        _set_config_path(
            config,
            path=path,
            value=value,
            source="cli.missingness_override",
            events=events,
        )


def _apply_default_cuda_fixed_layout_target_floor(
    config: GeneratorConfig,
    *,
    hw: HardwareInfo,
    events: list[ResolutionEvent],
) -> None:
    """Fill the fixed-layout target from the default CUDA floor when the config leaves it unset."""

    target_floor, _ = resolve_cuda_fixed_layout_target_cells_limits(hw)
    if target_floor is None:
        return
    current_target = config.runtime.fixed_layout_target_cells
    if current_target is not None:
        return
    _set_config_path(
        config,
        path="runtime.fixed_layout_target_cells",
        value=int(target_floor),
        source=_DEFAULT_CUDA_FIXED_LAYOUT_TARGET_SOURCE,
        events=events,
    )


def _apply_rows_override(
    config: GeneratorConfig,
    *,
    rows: object | None,
    source: str,
    events: list[ResolutionEvent],
) -> None:
    """Apply generate rows override."""

    if rows is None:
        return
    _set_config_path(
        config,
        path="dataset.rows",
        value=rows,
        source=source,
        events=events,
    )


def _apply_smoke_caps(
    config: GeneratorConfig,
    *,
    smoke_caps: BenchmarkSmokeCaps,
    events: list[ResolutionEvent],
) -> None:
    """Apply benchmark smoke caps and trace each field-level cap event."""

    cap_specs = (
        ("dataset.n_train", int(smoke_caps.n_train)),
        ("dataset.n_test", int(smoke_caps.n_test)),
        ("dataset.n_features_min", int(smoke_caps.n_features)),
        ("dataset.n_features_max", int(smoke_caps.n_features)),
        ("graph.n_nodes_min", int(smoke_caps.n_nodes)),
        ("graph.n_nodes_max", int(smoke_caps.n_nodes)),
    )
    for path, cap_value in cap_specs:
        parts = path.split(".")
        target: Any = config
        for part in parts[:-1]:
            target = getattr(target, part)
        field_name = parts[-1]
        old_value = int(getattr(target, field_name))
        new_value = min(old_value, cap_value)
        if old_value == new_value:
            continue
        setattr(target, field_name, new_value)
        _append_event(
            events=events,
            path=path,
            source="benchmark.suite_smoke_caps",
            old_value=old_value,
            new_value=new_value,
        )


def cap_rows_spec_to_total(config: GeneratorConfig, *, total_rows_cap: int) -> None:
    """Cap ``dataset.rows`` to ``total_rows_cap`` while preserving public row modes."""

    if int(total_rows_cap) < int(DATASET_ROWS_MIN_TOTAL):
        config.dataset.rows = None
        return

    normalized_rows = normalize_dataset_rows(config.dataset.rows)
    if normalized_rows is None:
        return

    if normalized_rows.mode == "fixed":
        assert normalized_rows.value is not None
        config.dataset.rows = DatasetRowsSpec(
            mode="fixed",
            value=min(int(normalized_rows.value), int(total_rows_cap)),
        )
        return
    if normalized_rows.mode == "range":
        assert normalized_rows.start is not None and normalized_rows.stop is not None
        capped_start = min(int(normalized_rows.start), int(total_rows_cap))
        capped_stop = min(int(normalized_rows.stop), int(total_rows_cap))
        if capped_start >= capped_stop:
            config.dataset.rows = DatasetRowsSpec(mode="fixed", value=capped_stop)
            return
        config.dataset.rows = DatasetRowsSpec(mode="range", start=capped_start, stop=capped_stop)
        return

    capped_choices = sorted(
        {min(int(choice), int(total_rows_cap)) for choice in normalized_rows.choices}
    )
    if len(capped_choices) == 1:
        config.dataset.rows = DatasetRowsSpec(mode="fixed", value=capped_choices[0])
        return
    config.dataset.rows = DatasetRowsSpec(mode="choices", choices=capped_choices)


def _apply_request_smoke_profile(
    config: GeneratorConfig,
    *,
    smoke_profile: RequestSmokeProfile,
    events: list[ResolutionEvent],
) -> None:
    """Apply request smoke profile overrides."""

    smoke_specs = (
        ("dataset.n_train", int(smoke_profile.n_train)),
        ("dataset.n_test", int(smoke_profile.n_test)),
        ("dataset.n_features_min", int(smoke_profile.n_features_min)),
        ("dataset.n_features_max", int(smoke_profile.n_features_max)),
        ("graph.n_nodes_min", int(smoke_profile.n_nodes_min)),
        ("graph.n_nodes_max", int(smoke_profile.n_nodes_max)),
    )
    for path, value in smoke_specs:
        _set_config_path(
            config,
            path=path,
            value=value,
            source="request.profile_smoke",
            events=events,
        )


def _apply_request_missingness_profile(
    config: GeneratorConfig,
    *,
    missingness_profile: str,
    events: list[ResolutionEvent],
) -> None:
    """Apply missingness profile defaults without leaking preset-local fields."""

    if missingness_profile == MISSINGNESS_MECHANISM_NONE:
        overlay = load_packaged_generator_config(_REQUEST_BASE_CONFIG_RESOURCE)
    else:
        overlay_resource = _REQUEST_MISSINGNESS_PRESET_RESOURCES[missingness_profile]
        overlay = load_packaged_generator_config(overlay_resource)

    for path in _REQUEST_MISSINGNESS_DATASET_PATHS:
        target = config.dataset
        source = overlay.dataset
        field_name = path.split(".")[-1]
        old_value = getattr(target, field_name)
        new_value = getattr(source, field_name)
        if old_value == new_value:
            continue
        setattr(target, field_name, new_value)
        _append_event(
            events=events,
            path=path,
            source="request.missingness_profile",
            old_value=old_value,
            new_value=new_value,
        )


def resolve_generate_config(
    config: GeneratorConfig,
    *,
    device_override: str | None,
    rows: object | None,
    rows_source: str = "cli.rows",
    hardware_policy: str,
    missing_rate: float | None,
    missing_mechanism: str | None,
    missing_mar_observed_fraction: float | None,
    missing_mar_logit_scale: float | None,
    missing_mnar_logit_scale: float | None,
    diagnostics_enabled: bool,
    post_policy_hook: Callable[[GeneratorConfig, list[ResolutionEvent]], None] | None = None,
) -> ResolvedGenerateConfig:
    """Resolve effective config for one generate command invocation."""

    resolved = _clone_config(config)
    trace_events: list[ResolutionEvent] = []

    requested_device = _normalize_requested_device(device_override, resolved.runtime.device)
    _set_config_path(
        resolved,
        path="runtime.device",
        value=requested_device,
        source="cli.device",
        events=trace_events,
    )

    hw = detect_hardware(requested_device)
    before_policy = resolved.to_dict()
    resolved = apply_hardware_policy(
        resolved,
        hw,
        policy_name=hardware_policy,
        validate=False,
    )
    after_policy = resolved.to_dict()
    _append_diff_events(
        before_policy,
        after_policy,
        source=f"hardware_policy.{str(hardware_policy).strip().lower()}",
        events=trace_events,
    )
    _apply_default_cuda_fixed_layout_target_floor(
        resolved,
        hw=hw,
        events=trace_events,
    )
    if post_policy_hook is not None:
        post_policy_hook(resolved, trace_events)

    _apply_rows_override(
        resolved,
        rows=rows,
        source=rows_source,
        events=trace_events,
    )

    _apply_missingness_overrides(
        resolved,
        missing_rate=missing_rate,
        missing_mechanism=missing_mechanism,
        missing_mar_observed_fraction=missing_mar_observed_fraction,
        missing_mar_logit_scale=missing_mar_logit_scale,
        missing_mnar_logit_scale=missing_mnar_logit_scale,
        events=trace_events,
    )

    if diagnostics_enabled:
        _set_config_path(
            resolved,
            path="diagnostics.enabled",
            value=True,
            source="cli.diagnostics",
            events=trace_events,
        )

    resolved.validate_generation_constraints()
    return ResolvedGenerateConfig(
        config=resolved,
        hardware=hw,
        requested_device=requested_device,
        trace_events=trace_events,
    )


def resolve_request_config(
    *,
    request: RequestFileConfig,
    device_override: str | None,
    hardware_policy: str,
) -> ResolvedRequestConfig:
    """Resolve effective config for one request-file execution."""

    base_config = load_packaged_generator_config(_REQUEST_BASE_CONFIG_RESOURCE)
    trace_events: list[ResolutionEvent] = []
    rows_override: object | None = request.rows
    rows_source = "request.rows"

    smoke_profile: RequestSmokeProfile | None = None
    if request.profile == REQUEST_PROFILE_SMOKE:
        smoke_profile = RequestSmokeProfile(
            n_train=128,
            n_test=32,
            n_features_min=8,
            n_features_max=12,
            n_nodes_min=2,
            n_nodes_max=12,
        )
        _apply_request_smoke_profile(
            base_config,
            smoke_profile=smoke_profile,
            events=trace_events,
        )
        rows_override = None

    _set_config_path(
        base_config,
        path="dataset.task",
        value=request.task,
        source="request.task",
        events=trace_events,
    )
    if request.seed is not None:
        _set_config_path(
            base_config,
            path="seed",
            value=int(request.seed),
            source="request.seed",
            events=trace_events,
        )
    if smoke_profile is not None:
        _set_config_path(
            base_config,
            path="dataset.rows",
            value=request.rows,
            source="request.rows",
            events=trace_events,
        )
    _apply_request_missingness_profile(
        base_config,
        missingness_profile=request.missingness_profile,
        events=trace_events,
    )
    _set_config_path(
        base_config,
        path="output.out_dir",
        value=str(Path(request.output_root) / "generated"),
        source="request.output_root",
        events=trace_events,
    )

    post_policy_hook: Callable[[GeneratorConfig, list[ResolutionEvent]], None] | None = None
    if smoke_profile is not None:

        def _reapply_request_smoke_profile(
            config: GeneratorConfig,
            events: list[ResolutionEvent],
        ) -> None:
            _apply_request_smoke_profile(
                config,
                smoke_profile=smoke_profile,
                events=events,
            )

        post_policy_hook = _reapply_request_smoke_profile

    resolved = resolve_generate_config(
        base_config,
        device_override=device_override,
        rows=rows_override,
        rows_source=rows_source,
        hardware_policy=hardware_policy,
        missing_rate=None,
        missing_mechanism=None,
        missing_mar_observed_fraction=None,
        missing_mar_logit_scale=None,
        missing_mnar_logit_scale=None,
        diagnostics_enabled=False,
        post_policy_hook=post_policy_hook,
    )
    return ResolvedRequestConfig(
        request=request,
        config=resolved.config,
        hardware=resolved.hardware,
        requested_device=resolved.requested_device,
        trace_events=trace_events + list(resolved.trace_events),
    )


def resolve_benchmark_preset_config(
    *,
    preset_key: str,
    config: GeneratorConfig,
    preset_device: str | None,
    suite: str,
    hardware_policy: str,
    smoke_caps: BenchmarkSmokeCaps | None,
) -> ResolvedBenchmarkPresetConfig:
    """Resolve effective config for one benchmark preset run."""

    resolved = _clone_config(config)
    trace_events: list[ResolutionEvent] = []

    requested_device = _normalize_requested_device(preset_device, resolved.runtime.device)
    _set_config_path(
        resolved,
        path="runtime.device",
        value=requested_device,
        source="benchmark.preset_device",
        events=trace_events,
    )

    hw = detect_hardware(requested_device)
    before_policy = resolved.to_dict()
    resolved = apply_hardware_policy(resolved, hw, policy_name=hardware_policy)
    after_policy = resolved.to_dict()
    _append_diff_events(
        before_policy,
        after_policy,
        source=f"hardware_policy.{str(hardware_policy).strip().lower()}",
        events=trace_events,
    )
    _apply_default_cuda_fixed_layout_target_floor(
        resolved,
        hw=hw,
        events=trace_events,
    )

    normalized_suite = str(suite).strip().lower()
    if normalized_suite == "smoke":
        if smoke_caps is None:
            raise ValueError("Benchmark smoke suite config resolution requires smoke cap values.")
        _apply_smoke_caps(resolved, smoke_caps=smoke_caps, events=trace_events)

    resolved.validate_generation_constraints()
    return ResolvedBenchmarkPresetConfig(
        preset_key=preset_key,
        config=resolved,
        hardware=hw,
        requested_device=requested_device,
        trace_events=trace_events,
    )


def serialize_resolution_events(events: list[ResolutionEvent]) -> list[dict[str, Any]]:
    """Convert trace dataclasses into JSON/YAML-safe dictionaries."""

    return [asdict(event) for event in events]


def append_config_diff_events(
    before: GeneratorConfig,
    after: GeneratorConfig,
    *,
    source: str,
    events: list[ResolutionEvent],
) -> None:
    """Append trace events describing config differences between two states."""

    _append_diff_events(
        before.to_dict(),
        after.to_dict(),
        source=source,
        events=events,
    )


__all__ = [
    "BenchmarkSmokeCaps",
    "RequestSmokeProfile",
    "ResolutionEvent",
    "ResolvedBenchmarkPresetConfig",
    "ResolvedGenerateConfig",
    "ResolvedRequestConfig",
    "append_config_diff_events",
    "cap_rows_spec_to_total",
    "resolve_benchmark_preset_config",
    "resolve_generate_config",
    "resolve_request_config",
    "serialize_resolution_events",
]



from __future__ import annotations

import pytest
import yaml

from dagzoo.config import (
    MISSINGNESS_MECHANISM_MAR,
    MISSINGNESS_MECHANISM_MCAR,
    MISSINGNESS_MECHANISM_MNAR,
    MISSINGNESS_MECHANISM_NONE,
    REQUEST_FILE_VERSION_V1,
    REQUEST_PROFILE_DEFAULT,
    REQUEST_PROFILE_SMOKE,
    REQUEST_TASK_CLASSIFICATION,
    REQUEST_TASK_REGRESSION,
    RequestFileConfig,
)


def test_request_file_accepts_minimal_classification_payload() -> None:
    cfg = RequestFileConfig.from_dict(
        {
            "version": REQUEST_FILE_VERSION_V1,
            "task": REQUEST_TASK_CLASSIFICATION,
            "dataset_count": 25,
            "rows": 1024,
            "profile": REQUEST_PROFILE_DEFAULT,
            "output_root": "requests/corpus_default",
        }
    )

    assert cfg.version == REQUEST_FILE_VERSION_V1
    assert cfg.task == REQUEST_TASK_CLASSIFICATION
    assert cfg.dataset_count == 25
    assert cfg.rows.mode == "fixed"
    assert cfg.rows.value == 1024
    assert cfg.profile == REQUEST_PROFILE_DEFAULT
    assert cfg.output_root == "requests/corpus_default"
    assert cfg.missingness_profile == MISSINGNESS_MECHANISM_NONE
    assert cfg.seed is None


def test_request_file_accepts_range_rows_smoke_profile_and_seed() -> None:
    cfg = RequestFileConfig.from_dict(
        {
            "version": "V1",
            "task": REQUEST_TASK_REGRESSION,
            "dataset_count": 3,
            "rows": "1024..4096",
            "profile": "SMOKE",
            "missingness_profile": "mcar",
            "output_root": "requests/regression_smoke",
            "seed": 42,
        }
    )

    assert cfg.version == REQUEST_FILE_VERSION_V1
    assert cfg.task == REQUEST_TASK_REGRESSION
    assert cfg.rows.mode == "range"
    assert cfg.rows.start == 1024
    assert cfg.rows.stop == 4096
    assert cfg.profile == REQUEST_PROFILE_SMOKE
    assert cfg.missingness_profile == MISSINGNESS_MECHANISM_MCAR
    assert cfg.seed == 42


def test_request_file_accepts_choice_rows_and_each_missingness_profile() -> None:
    for missingness_profile in (
        MISSINGNESS_MECHANISM_NONE,
        MISSINGNESS_MECHANISM_MCAR,
        MISSINGNESS_MECHANISM_MAR,
        MISSINGNESS_MECHANISM_MNAR,
    ):
        cfg = RequestFileConfig.from_dict(
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 8,
                "rows": "1024,2048,4096",
                "profile": REQUEST_PROFILE_DEFAULT,
                "missingness_profile": missingness_profile,
                "output_root": "requests/profile_matrix",
            }
        )
        assert cfg.rows.mode == "choices"
        assert cfg.rows.choices == [1024, 2048, 4096]
        assert cfg.missingness_profile == missingness_profile


def test_request_file_from_yaml_loads_top_level_mapping(tmp_path) -> None:
    path = tmp_path / "request.yaml"
    path.write_text(
        yaml.safe_dump(
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 10,
                "rows": 2048,
                "profile": REQUEST_PROFILE_DEFAULT,
                "missingness_profile": MISSINGNESS_MECHANISM_MAR,
                "output_root": "requests/mar_default",
            },
            sort_keys=False,
        ),
        encoding="utf-8",
    )

    cfg = RequestFileConfig.from_yaml(path)

    assert cfg.task == REQUEST_TASK_CLASSIFICATION
    assert cfg.dataset_count == 10
    assert cfg.rows.mode == "fixed"
    assert cfg.rows.value == 2048
    assert cfg.missingness_profile == MISSINGNESS_MECHANISM_MAR


def test_request_file_from_yaml_rejects_non_string_top_level_keys(tmp_path) -> None:
    path = tmp_path / "request.yaml"
    path.write_text(
        yaml.safe_dump(
            {
                1: "unexpected",
                "runtime": {"device": "cpu"},
            },
            sort_keys=False,
        ),
        encoding="utf-8",
    )

    with pytest.raises(
        ValueError,
        match=r"Request-file field names must be strings",
    ):
        RequestFileConfig.from_yaml(path)


@pytest.mark.parametrize(
    "rows_value",
    [
        [1024, 2048],
        {"mode": "fixed", "value": 1024},
    ],
)
def test_request_file_from_yaml_rejects_non_public_rows_encodings(
    tmp_path, rows_value: object
) -> None:
    path = tmp_path / "request.yaml"
    path.write_text(
        yaml.safe_dump(
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 2,
                "rows": rows_value,
                "profile": REQUEST_PROFILE_DEFAULT,
                "output_root": "requests/out",
            },
            sort_keys=False,
        ),
        encoding="utf-8",
    )

    with pytest.raises(
        ValueError,
        match=r"rows must use one of the public request-file encodings",
    ):
        RequestFileConfig.from_yaml(path)


@pytest.mark.parametrize(
    ("payload", "expected"),
    [
        (
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 25,
                "rows": 1024,
                "profile": REQUEST_PROFILE_DEFAULT,
                "output_root": "requests/corpus_default",
            },
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 25,
                "rows": 1024,
                "profile": REQUEST_PROFILE_DEFAULT,
                "output_root": "requests/corpus_default",
                "missingness_profile": MISSINGNESS_MECHANISM_NONE,
            },
        ),
        (
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_REGRESSION,
                "dataset_count": 3,
                "rows": "1024..4096",
                "profile": REQUEST_PROFILE_SMOKE,
                "missingness_profile": MISSINGNESS_MECHANISM_MCAR,
                "output_root": "requests/regression_smoke",
                "seed": 42,
            },
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_REGRESSION,
                "dataset_count": 3,
                "rows": "1024..4096",
                "profile": REQUEST_PROFILE_SMOKE,
                "output_root": "requests/regression_smoke",
                "missingness_profile": MISSINGNESS_MECHANISM_MCAR,
                "seed": 42,
            },
        ),
        (
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 8,
                "rows": "1024,2048,4096",
                "profile": REQUEST_PROFILE_DEFAULT,
                "missingness_profile": MISSINGNESS_MECHANISM_MAR,
                "output_root": "requests/profile_matrix",
            },
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 8,
                "rows": "1024,2048,4096",
                "profile": REQUEST_PROFILE_DEFAULT,
                "output_root": "requests/profile_matrix",
                "missingness_profile": MISSINGNESS_MECHANISM_MAR,
            },
        ),
    ],
)
def test_request_file_to_dict_emits_public_wire_shape(
    payload: dict[str, object], expected: dict[str, object]
) -> None:
    cfg = RequestFileConfig.from_dict(payload)

    serialized = cfg.to_dict()

    assert serialized == expected
    assert RequestFileConfig.from_dict(serialized).to_dict() == expected


def test_request_file_to_dict_omits_unset_seed() -> None:
    cfg = RequestFileConfig.from_dict(
        {
            "version": REQUEST_FILE_VERSION_V1,
            "task": REQUEST_TASK_CLASSIFICATION,
            "dataset_count": 25,
            "rows": 1024,
            "profile": REQUEST_PROFILE_DEFAULT,
            "output_root": "requests/corpus_default",
        }
    )

    assert "seed" not in cfg.to_dict()


@pytest.mark.parametrize("rows_value", [300, "300..600"])
def test_request_file_rows_errors_use_public_field_names(rows_value: object) -> None:
    with pytest.raises(ValueError) as exc_info:
        RequestFileConfig.from_dict(
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 1,
                "rows": rows_value,
                "profile": REQUEST_PROFILE_DEFAULT,
                "output_root": "requests/out",
            }
        )

    message = str(exc_info.value)
    assert "rows" in message
    assert "dataset.rows" not in message


@pytest.mark.parametrize(
    ("payload", "pattern"),
    [
        (
            {
                "version": "v2",
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 1,
                "rows": 1024,
                "profile": REQUEST_PROFILE_DEFAULT,
                "output_root": "requests/out",
            },
            r"version must be one of: v1",
        ),
        (
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": "clustering",
                "dataset_count": 1,
                "rows": 1024,
                "profile": REQUEST_PROFILE_DEFAULT,
                "output_root": "requests/out",
            },
            r"task must be one of: classification, regression",
        ),
        (
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 0,
                "rows": 1024,
                "profile": REQUEST_PROFILE_DEFAULT,
                "output_root": "requests/out",
            },
            r"dataset_count must be a positive integer",
        ),
        (
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 1,
                "rows": "1024",
                "profile": REQUEST_PROFILE_DEFAULT,
                "output_root": "requests/out",
            },
            r"rows must use one of the public request-file encodings",
        ),
        (
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 1,
                "rows": "300..600",
                "profile": REQUEST_PROFILE_DEFAULT,
                "output_root": "requests/out",
            },
            r"rows\.start",
        ),
        (
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 1,
                "rows": 1024,
                "profile": "benchmark_cpu",
                "output_root": "requests/out",
            },
            r"profile must be one of: default, smoke",
        ),
        (
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 1,
                "rows": 1024,
                "profile": REQUEST_PROFILE_DEFAULT,
                "missingness_profile": "custom",
                "output_root": "requests/out",
            },
            r"missingness_profile must be one of: mar, mcar, mnar, none",
        ),
        (
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 1,
                "rows": 1024,
                "profile": REQUEST_PROFILE_DEFAULT,
                "output_root": "requests/out",
                "seed": 4294967296,
            },
            r"seed must be an integer in \[0, 4294967295\]",
        ),
        (
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 1,
                "rows": 1024,
                "profile": REQUEST_PROFILE_DEFAULT,
                "output_root": "",
            },
            r"output_root must be a non-empty string",
        ),
    ],
)
def test_request_file_rejects_invalid_values(payload: dict[str, object], pattern: str) -> None:
    with pytest.raises(ValueError, match=pattern):
        RequestFileConfig.from_dict(payload)


@pytest.mark.parametrize(
    "field_name", ["runtime", "filter", "graph", "dataset", "n_train", "n_test"]
)
def test_request_file_rejects_internal_config_fields(field_name: str) -> None:
    payload = {
        "version": REQUEST_FILE_VERSION_V1,
        "task": REQUEST_TASK_CLASSIFICATION,
        "dataset_count": 1,
        "rows": 1024,
        "profile": REQUEST_PROFILE_DEFAULT,
        "output_root": "requests/out",
        field_name: {},
    }
    if field_name in {"n_train", "n_test"}:
        payload[field_name] = 128

    with pytest.raises(
        ValueError,
        match=r"is not part of the public request-file contract",
    ):
        RequestFileConfig.from_dict(payload)


def test_request_file_rejects_unknown_extra_fields() -> None:
    with pytest.raises(ValueError, match=r"Unknown request-file field: 'unexpected'"):
        RequestFileConfig.from_dict(
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 1,
                "rows": 1024,
                "profile": REQUEST_PROFILE_DEFAULT,
                "output_root": "requests/out",
                "unexpected": "value",
            }
        )


def test_request_file_rejects_missing_required_fields() -> None:
    with pytest.raises(ValueError, match=r"Missing required request-file field: 'rows'"):
        RequestFileConfig.from_dict(
            {
                "version": REQUEST_FILE_VERSION_V1,
                "task": REQUEST_TASK_CLASSIFICATION,
                "dataset_count": 1,
                "profile": REQUEST_PROFILE_DEFAULT,
                "output_root": "requests/out",
            }
        )



"""Request-file execution helpers."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from time import perf_counter

from dagzoo.config import REQUEST_PROFILE_SMOKE, RequestFileConfig, clone_generator_config
from dagzoo.core.config_resolution import (
    append_config_diff_events,
    cap_rows_spec_to_total,
    resolve_request_config,
    serialize_resolution_events,
)
from dagzoo.core.fixed_layout.runtime import realize_generation_config_for_run
from dagzoo.core.request_handoff import (
    REQUEST_HANDOFF_MANIFEST_FILENAME,
    write_request_handoff_manifest,
)
from dagzoo.filtering import DeferredFilterRunResult
from dagzoo.filtering.deferred_filter import MANIFEST_FILENAME, SUMMARY_FILENAME
from dagzoo.io.parquet_writer import write_packed_parquet_shards_stream

from .common import get_cli_public_api
from .effective_config import (
    print_effective_config,
    print_resolution_trace,
    write_effective_config,
    write_effective_config_trace,
)


@dataclass(slots=True)
class RequestRunResult:
    """Result payload for one request execution."""

    request_path: Path
    run_root: Path
    generated_dir: Path
    filter_dir: Path
    curated_dir: Path
    effective_config_path: Path
    effective_config_trace_path: Path
    handoff_manifest_path: Path
    generated_datasets: int
    filter_result: DeferredFilterRunResult


def _request_output_dirs(run_root: Path) -> tuple[Path, Path, Path]:
    """Return the fixed request-run artifact directories."""

    return run_root / "generated", run_root / "filter", run_root / "curated"


def _ensure_request_run_output_safe(run_root: Path) -> tuple[Path, Path, Path]:
    """Fail fast when one request-run output root already holds prior artifacts."""

    if run_root.exists() and not run_root.is_dir():
        raise RuntimeError(f"Request output_root must be a directory path: {run_root}")

    generated_dir, filter_dir, curated_dir = _request_output_dirs(run_root)
    for path in (generated_dir, filter_dir, curated_dir):
        if path.exists() and not path.is_dir():
            raise RuntimeError(f"Request artifact path must be a directory: {path}")

    stale_generated = next(generated_dir.glob("shard_*"), None) if generated_dir.exists() else None
    if stale_generated is not None:
        raise RuntimeError(
            "Request generated output already contains shard data: "
            f"{generated_dir}. Remove existing shard_* folders or choose a new output_root."
        )

    if filter_dir.exists() and (
        (filter_dir / MANIFEST_FILENAME).exists() or (filter_dir / SUMMARY_FILENAME).exists()
    ):
        raise RuntimeError(
            "Request filter output already contains prior artifacts: "
            f"{filter_dir}. Remove prior filter artifacts or choose a new output_root."
        )

    stale_curated = next(curated_dir.glob("shard_*"), None) if curated_dir.exists() else None
    if stale_curated is not None:
        raise RuntimeError(
            "Request curated output already contains shard data: "
            f"{curated_dir}. Remove existing shard_* folders or choose a new output_root."
        )

    handoff_manifest_path = run_root / REQUEST_HANDOFF_MANIFEST_FILENAME
    if handoff_manifest_path.exists():
        raise RuntimeError(
            "Request output_root already contains a prior handoff manifest: "
            f"{handoff_manifest_path}. Remove the existing manifest or choose a new output_root."
        )

    return generated_dir, filter_dir, curated_dir


def run_request_execution(
    *,
    request_path: str | Path,
    device_override: str | None,
    hardware_policy: str,
    n_jobs_override: int | None,
    print_effective_config_flag: bool,
    print_resolution_trace_flag: bool,
) -> RequestRunResult:
    """Resolve and execute one request-file run through canonical generate->filter."""

    request_file = RequestFileConfig.from_yaml(request_path)
    run_root = Path(request_file.output_root)
    generated_dir, filter_dir, curated_dir = _ensure_request_run_output_safe(run_root)
    resolved = resolve_request_config(
        request=request_file,
        device_override=device_override,
        hardware_policy=hardware_policy,
    )
    trace_events = list(resolved.trace_events)
    pre_realization_config = clone_generator_config(resolved.config, revalidate=False)
    if request_file.profile == REQUEST_PROFILE_SMOKE:
        cap_rows_spec_to_total(
            pre_realization_config,
            total_rows_cap=int(
                pre_realization_config.dataset.n_train + pre_realization_config.dataset.n_test
            ),
        )
        append_config_diff_events(
            resolved.config,
            pre_realization_config,
            source="request.smoke_rows_cap",
            events=trace_events,
        )

    seed = pre_realization_config.seed
    config, run_seed, requested_device, resolved_device = realize_generation_config_for_run(
        pre_realization_config,
        seed=seed,
        device=resolved.requested_device,
    )
    if bool(config.filter.enabled):
        raise ValueError(
            "Inline filtering has been removed from request execution. "
            "Request runs must resolve to filter.enabled=false and use deferred filtering."
        )

    append_config_diff_events(
        pre_realization_config,
        config,
        source="request.run_realization",
        events=trace_events,
    )
    trace_payload = serialize_resolution_events(trace_events)

    if print_effective_config_flag:
        print_effective_config(config, header="Effective config:")

    effective_config_path = write_effective_config(config, generated_dir / "effective_config.yaml")
    effective_config_trace_path = write_effective_config_trace(
        trace_payload,
        generated_dir / "effective_config_trace.yaml",
    )
    if print_resolution_trace_flag:
        print_resolution_trace(trace_payload, header="Resolution trace:")

    cli_api = get_cli_public_api()
    print(
        f"Hardware backend={resolved.hardware.backend} device='{resolved.hardware.device_name}' "
        f"memory_gb={resolved.hardware.total_memory_gb} "
        f"peak_flops={resolved.hardware.peak_flops:.3e} tier={resolved.hardware.tier} "
        f"hardware_policy={hardware_policy}"
    )
    generation_started_at = perf_counter()
    stream = cli_api.generate_batch_iter(
        config,
        num_datasets=request_file.dataset_count,
        seed=run_seed,
        device=requested_device,
    )
    written = write_packed_parquet_shards_stream(
        stream,
        out_dir=generated_dir,
        shard_size=config.output.shard_size,
        compression=config.output.compression,
    )
    generation_elapsed_seconds = perf_counter() - generation_started_at
    filter_started_at = perf_counter()
    filter_result = cli_api.run_deferred_filter(
        in_dir=generated_dir,
        out_dir=filter_dir,
        curated_out_dir=curated_dir,
        n_jobs_override=n_jobs_override,
    )
    filter_elapsed_seconds = perf_counter() - filter_started_at
    handoff_manifest_path = write_request_handoff_manifest(
        request_path=request_path,
        request=request_file,
        run_root=run_root,
        generated_dir=generated_dir,
        filter_dir=filter_dir,
        filtered_corpus_dir=curated_dir,
        effective_config_path=effective_config_path,
        effective_config_trace_path=effective_config_trace_path,
        filter_manifest_path=filter_result.manifest_path,
        filter_summary_path=filter_result.summary_path,
        generated_datasets=int(written),
        generation_elapsed_seconds=float(generation_elapsed_seconds),
        filter_total_datasets=int(filter_result.total_datasets),
        filter_accepted_datasets=int(filter_result.accepted_datasets),
        filter_rejected_datasets=int(filter_result.rejected_datasets),
        filter_elapsed_seconds=float(filter_elapsed_seconds),
        requested_device=str(resolved.requested_device),
        resolved_device=str(resolved_device),
        hardware_backend=str(resolved.hardware.backend),
        hardware_device_name=str(resolved.hardware.device_name),
        hardware_tier=str(resolved.hardware.tier),
        hardware_policy=str(hardware_policy),
    )

    return RequestRunResult(
        request_path=Path(request_path),
        run_root=run_root,
        generated_dir=generated_dir,
        filter_dir=filter_dir,
        curated_dir=curated_dir,
        effective_config_path=effective_config_path,
        effective_config_trace_path=effective_config_trace_path,
        handoff_manifest_path=handoff_manifest_path,
        generated_datasets=int(written),
        filter_result=filter_result,
    )



"""Dataset filters."""

from .deferred_filter import DeferredFilterRunResult, run_deferred_filter
from .extra_trees_filter import apply_extra_trees_filter

__all__ = ["DeferredFilterRunResult", "apply_extra_trees_filter", "run_deferred_filter"]



"""Helpers for benchmark stage-level throughput measurement."""

from __future__ import annotations

import tempfile
import time
from collections.abc import Callable, Iterable, Sequence
from dataclasses import dataclass, field
from typing import Any

import numpy as np

from dagzoo.bench.constants import SECONDS_PER_MINUTE
from dagzoo.config import GeneratorConfig
from dagzoo.filtering.extra_trees_filter import _apply_extra_trees_filter_numpy
from dagzoo.io.parquet_writer import write_packed_parquet_shards_stream
from dagzoo.math import coerce_optional_finite_float as _coerce_optional_finite_float
from dagzoo.math import to_numpy as _to_numpy
from dagzoo.rng import SEED32_MAX
from dagzoo.types import DatasetBundle


@dataclass(slots=True)
class StageSampleCollector:
    """Collect a bounded sample of emitted bundles for stage timing probes."""

    max_samples: int
    bundles: list[DatasetBundle] = field(default_factory=list)

    def update(self, bundle: DatasetBundle) -> None:
        """Store one emitted bundle if the sample has remaining capacity."""

        if len(self.bundles) >= max(0, int(self.max_samples)):
            return
        self.bundles.append(bundle)


@dataclass(slots=True)
class FilterStageMeasurement:
    """Deferred filter stage replay metrics over sampled bundles."""

    datasets_per_minute: float
    filter_attempts_total: int
    filter_accepted_datasets: int
    filter_rejections_total: int
    filter_rejected_datasets: int
    accepted_true_fraction: float | None = None
    wins_ratio_mean: float | None = None
    threshold_effective_mean: float | None = None
    threshold_delta_mean: float | None = None
    n_valid_oob_mean: float | None = None
    reason_counts: dict[str, int] = field(default_factory=dict)
    accepted_bundles: list[DatasetBundle] = field(default_factory=list)


def measure_write_datasets_per_minute(
    bundles: Sequence[DatasetBundle],
    *,
    config: GeneratorConfig,
) -> float:
    """Measure parquet write throughput on sampled bundles."""

    num_bundles = len(bundles)
    if num_bundles <= 0:
        return 0.0

    start = time.perf_counter()
    with tempfile.TemporaryDirectory(prefix="dagzoo_stage_write_") as tmp_dir:
        _ = write_packed_parquet_shards_stream(
            bundles,
            out_dir=tmp_dir,
            shard_size=max(1, int(config.output.shard_size)),
            compression=str(config.output.compression),
        )
    elapsed = time.perf_counter() - start
    if elapsed <= 0.0:
        return 0.0
    return (float(num_bundles) / elapsed) * SECONDS_PER_MINUTE


def _coerce_bundle_seed(bundle: DatasetBundle, *, fallback_seed: int) -> int:
    """Resolve per-bundle seed for deterministic deferred filter replay."""

    metadata = bundle.metadata
    for key in ("dataset_seed", "seed"):
        raw_seed: Any = metadata.get(key)
        if isinstance(raw_seed, bool):
            raw_seed = None
        if isinstance(raw_seed, float):
            if raw_seed.is_integer():
                raw_seed = int(raw_seed)
            else:
                raw_seed = None
        if isinstance(raw_seed, int):
            return int(raw_seed % (SEED32_MAX + 1))
    return int(fallback_seed % (SEED32_MAX + 1))


def _mean_or_none(values: list[float]) -> float | None:
    """Return the mean of collected values or ``None`` when empty."""

    if not values:
        return None
    return float(sum(values) / len(values))


def replay_filter_stage_metrics(
    bundles: Iterable[DatasetBundle],
    *,
    config: GeneratorConfig,
    on_accepted_bundle: Callable[[DatasetBundle], None] | None = None,
) -> FilterStageMeasurement:
    """Replay deferred filter stage over a bundle stream and return throughput + outcomes."""

    if not bool(config.filter.enabled):
        return FilterStageMeasurement(
            datasets_per_minute=0.0,
            filter_attempts_total=0,
            filter_accepted_datasets=0,
            filter_rejections_total=0,
            filter_rejected_datasets=0,
        )

    attempts_total = 0
    accepted_total = 0
    rejections_total = 0
    wins_ratio_values: list[float] = []
    threshold_effective_values: list[float] = []
    threshold_delta_values: list[float] = []
    n_valid_oob_values: list[float] = []
    reason_counts: dict[str, int] = {}
    start = time.perf_counter()
    callback_elapsed = 0.0
    for idx, bundle in enumerate(bundles):
        x_train = _to_numpy(bundle.X_train).astype("float32", copy=False)
        x_test = _to_numpy(bundle.X_test).astype("float32", copy=False)
        y_train = _to_numpy(bundle.y_train)
        y_test = _to_numpy(bundle.y_test)
        x_all = x_train
        if x_test.size > 0:
            x_all = np.concatenate([x_train, x_test], axis=0)
        y_all = y_train
        if y_test.size > 0:
            y_all = np.concatenate([y_train, y_test], axis=0)
        if str(config.dataset.task) == "classification":
            y_all = y_all.astype("int64", copy=False)
        else:
            y_all = y_all.astype("float32", copy=False)

        attempts_total += 1
        accepted, _details = _apply_extra_trees_filter_numpy(
            x_all,
            y_all,
            task=str(config.dataset.task),
            seed=_coerce_bundle_seed(bundle, fallback_seed=config.seed + idx),
            n_estimators=int(config.filter.n_estimators),
            max_depth=int(config.filter.max_depth),
            min_samples_leaf=int(config.filter.min_samples_leaf),
            max_leaf_nodes=config.filter.max_leaf_nodes,
            max_features=config.filter.max_features,
            n_bootstrap=int(config.filter.n_bootstrap),
            threshold=float(config.filter.threshold),
            n_jobs=int(config.filter.n_jobs),
        )
        if bool(accepted):
            accepted_total += 1
            if on_accepted_bundle is not None:
                callback_start = time.perf_counter()
                on_accepted_bundle(bundle)
                callback_elapsed += max(0.0, time.perf_counter() - callback_start)
        else:
            rejections_total += 1
        wins_ratio = _coerce_optional_finite_float(_details.get("wins_ratio"))
        if wins_ratio is not None:
            wins_ratio_values.append(float(wins_ratio))
        threshold_effective = _coerce_optional_finite_float(_details.get("threshold_effective"))
        if threshold_effective is not None:
            threshold_effective_values.append(float(threshold_effective))
        threshold_delta = _coerce_optional_finite_float(_details.get("threshold_delta"))
        if threshold_delta is not None:
            threshold_delta_values.append(float(threshold_delta))
        n_valid_oob = _coerce_optional_finite_float(_details.get("n_valid_oob"))
        if n_valid_oob is not None:
            n_valid_oob_values.append(float(n_valid_oob))
        reason = _details.get("reason")
        if isinstance(reason, str) and reason:
            reason_counts[reason] = reason_counts.get(reason, 0) + 1

    elapsed = max(0.0, time.perf_counter() - start - callback_elapsed)
    dpm = ((float(attempts_total) / elapsed) * SECONDS_PER_MINUTE) if elapsed > 0.0 else 0.0
    return FilterStageMeasurement(
        datasets_per_minute=float(dpm),
        filter_attempts_total=int(attempts_total),
        filter_accepted_datasets=int(accepted_total),
        filter_rejections_total=int(rejections_total),
        filter_rejected_datasets=int(rejections_total),
        accepted_true_fraction=(
            float(accepted_total) / float(attempts_total) if attempts_total > 0 else None
        ),
        wins_ratio_mean=_mean_or_none(wins_ratio_values),
        threshold_effective_mean=_mean_or_none(threshold_effective_values),
        threshold_delta_mean=_mean_or_none(threshold_delta_values),
        n_valid_oob_mean=_mean_or_none(n_valid_oob_values),
        reason_counts=dict(sorted(reason_counts.items())),
    )


def measure_filter_stage_metrics(
    bundles: Sequence[DatasetBundle],
    *,
    config: GeneratorConfig,
    collect_accepted_bundles: bool = False,
) -> FilterStageMeasurement:
    """Replay deferred filter stage over sampled bundles and return throughput + outcomes."""

    accepted_bundles: list[DatasetBundle] = []
    measurement = replay_filter_stage_metrics(
        bundles,
        config=config,
        on_accepted_bundle=accepted_bundles.append if collect_accepted_bundles else None,
    )
    if collect_accepted_bundles:
        measurement.accepted_bundles = accepted_bundles
    return measurement



import pytest

from dagzoo.config import (
    NOISE_FAMILY_GAUSSIAN,
    NOISE_FAMILY_LAPLACE,
    NOISE_FAMILY_MIXTURE,
    NOISE_FAMILY_STUDENT_T,
    GeneratorConfig,
)


def test_noise_config_defaults_from_default_yaml() -> None:
    cfg = GeneratorConfig.from_yaml("configs/default.yaml")
    assert cfg.noise.family == NOISE_FAMILY_GAUSSIAN
    assert cfg.noise.base_scale == pytest.approx(1.0)
    assert cfg.noise.student_t_df == pytest.approx(5.0)
    assert cfg.noise.mixture_weights is None


@pytest.mark.parametrize(
    "family",
    [
        NOISE_FAMILY_GAUSSIAN,
        NOISE_FAMILY_LAPLACE,
        NOISE_FAMILY_STUDENT_T,
        NOISE_FAMILY_MIXTURE,
    ],
)
def test_noise_config_accepts_supported_families(family: str) -> None:
    cfg = GeneratorConfig.from_dict({"noise": {"family": family}})
    assert cfg.noise.family == family


def test_noise_config_rejects_unknown_family() -> None:
    with pytest.raises(ValueError, match="Unsupported noise.family"):
        GeneratorConfig.from_dict({"noise": {"family": "exponential"}})


def test_noise_config_rejects_removed_legacy_family() -> None:
    with pytest.raises(ValueError, match="Unsupported noise.family"):
        GeneratorConfig.from_dict({"noise": {"family": "legacy"}})


def test_noise_config_rejects_student_t_df_at_or_below_two() -> None:
    with pytest.raises(ValueError, match=r"noise\.student_t_df must be a finite value > 2"):
        GeneratorConfig.from_dict({"noise": {"family": "student_t", "student_t_df": 2.0}})


def test_noise_config_rejects_mixture_weights_when_family_not_mixture() -> None:
    with pytest.raises(ValueError, match="noise.mixture_weights is only allowed"):
        GeneratorConfig.from_dict(
            {
                "noise": {
                    "family": "gaussian",
                    "mixture_weights": {"gaussian": 0.5, "laplace": 0.5},
                }
            }
        )


def test_noise_config_normalizes_mixture_weights() -> None:
    cfg = GeneratorConfig.from_dict(
        {
            "noise": {
                "family": "mixture",
                "mixture_weights": {"gaussian": 2.0, "laplace": 1.0, "student_t": 1.0},
            }
        }
    )
    weights = cfg.noise.mixture_weights
    assert weights is not None
    assert sum(weights.values()) == pytest.approx(1.0)
    assert weights["gaussian"] == pytest.approx(0.5)
    assert weights["laplace"] == pytest.approx(0.25)
    assert weights["student_t"] == pytest.approx(0.25)


def test_noise_config_rejects_mixture_weights_with_unknown_key() -> None:
    with pytest.raises(ValueError, match="Unsupported noise.mixture_weights key"):
        GeneratorConfig.from_dict(
            {
                "noise": {
                    "family": "mixture",
                    "mixture_weights": {"gaussian": 1.0, "exponential": 1.0},
                }
            }
        )


def test_noise_config_rejects_mixture_weights_with_nonpositive_total() -> None:
    with pytest.raises(ValueError, match="positive total weight"):
        GeneratorConfig.from_dict(
            {
                "noise": {
                    "family": "mixture",
                    "mixture_weights": {"gaussian": 0.0, "laplace": 0.0, "student_t": 0.0},
                }
            }
        )


def test_noise_config_stably_normalizes_large_mixture_weights() -> None:
    cfg = GeneratorConfig.from_dict(
        {
            "noise": {
                "family": "mixture",
                "mixture_weights": {"gaussian": 1e308, "laplace": 1e308},
            }
        }
    )
    weights = cfg.noise.mixture_weights
    assert weights is not None
    assert sum(weights.values()) == pytest.approx(1.0)
    assert weights["gaussian"] == pytest.approx(0.5)
    assert weights["laplace"] == pytest.approx(0.5)



"""Generate command handler."""

from __future__ import annotations

import argparse
from collections.abc import Iterator
from pathlib import Path
from typing import Any

from dagzoo.core.config_resolution import (
    append_config_diff_events,
    resolve_generate_config,
    serialize_resolution_events,
)
from dagzoo.core.fixed_layout.runtime import realize_generation_config_for_run
from dagzoo.diagnostics import (
    CoverageAggregator,
    write_coverage_summary_json,
    write_coverage_summary_markdown,
)
from dagzoo.diagnostics_targets import build_diagnostics_aggregation_config
from dagzoo.io.parquet_writer import write_packed_parquet_shards_stream

from ..common import get_cli_public_api, load_config_or_usage_error, raise_usage_error
from ..effective_config import (
    print_effective_config,
    print_resolution_trace,
    write_effective_config,
    write_effective_config_trace,
)


def _write_generate_diagnostics_artifacts(
    diagnostics_aggregator: CoverageAggregator,
    *,
    diagnostics_out_dir: Path,
) -> None:
    """Write generation diagnostics coverage artifacts and print output paths."""

    summary = diagnostics_aggregator.build_summary()
    json_path = write_coverage_summary_json(summary, diagnostics_out_dir / "coverage_summary.json")
    md_path = write_coverage_summary_markdown(summary, diagnostics_out_dir / "coverage_summary.md")
    print(f"Wrote diagnostics artifacts: {json_path} and {md_path}")


def run_generate_command(args: argparse.Namespace) -> int:
    """Execute the ``generate`` command."""

    cli_api = get_cli_public_api()
    config = load_config_or_usage_error(args.config)
    try:
        resolved = resolve_generate_config(
            config,
            device_override=args.device,
            rows=args.rows,
            hardware_policy=str(args.hardware_policy),
            missing_rate=args.missing_rate,
            missing_mechanism=args.missing_mechanism,
            missing_mar_observed_fraction=args.missing_mar_observed_fraction,
            missing_mar_logit_scale=args.missing_mar_logit_scale,
            missing_mnar_logit_scale=args.missing_mnar_logit_scale,
            diagnostics_enabled=bool(args.diagnostics),
        )
    except ValueError as exc:
        raise_usage_error(str(exc))

    seed = args.seed if args.seed is not None else resolved.config.seed
    config, run_seed, requested_device, _resolved_device = realize_generation_config_for_run(
        resolved.config,
        seed=seed,
        device=resolved.requested_device,
    )
    if bool(config.filter.enabled):
        raise_usage_error(
            "Inline filtering has been removed from generate. Set filter.enabled=false and run "
            "`dagzoo filter --in <shard_dir> --out <out_dir>` after generation."
        )
    hw = resolved.hardware
    trace_events = list(resolved.trace_events)
    append_config_diff_events(
        resolved.config,
        config,
        source="generate.run_realization",
        events=trace_events,
    )
    trace_payload = serialize_resolution_events(trace_events)
    out_dir = args.out or config.output.out_dir
    effective_config_root = out_dir
    if effective_config_root is None:
        effective_config_root = (
            args.diagnostics_out_dir or config.diagnostics.out_dir or "effective_config_artifacts"
        )
    if args.print_effective_config:
        print_effective_config(config, header="Effective config:")

    effective_config_path = write_effective_config(
        config,
        Path(effective_config_root) / "effective_config.yaml",
    )
    trace_path = write_effective_config_trace(
        trace_payload,
        Path(effective_config_root) / "effective_config_trace.yaml",
    )
    print(f"Wrote effective config: {effective_config_path}")
    print(f"Wrote effective config trace: {trace_path}")
    if args.print_resolution_trace:
        print_resolution_trace(trace_payload, header="Resolution trace:")

    diagnostics_enabled = bool(config.diagnostics.enabled)
    diagnostics_out_dir: Path | None = None
    diagnostics_aggregator: CoverageAggregator | None = None
    if diagnostics_enabled:
        diagnostics_root = args.diagnostics_out_dir or config.diagnostics.out_dir or out_dir
        if diagnostics_root is None:
            diagnostics_root = "diagnostics_artifacts"
        diagnostics_out_dir = Path(diagnostics_root)
        diagnostics_aggregator = cli_api.CoverageAggregator(
            build_diagnostics_aggregation_config(config.diagnostics)
        )
    print(
        f"Hardware backend={hw.backend} device='{hw.device_name}' "
        f"memory_gb={hw.total_memory_gb} peak_flops={hw.peak_flops:.3e} tier={hw.tier} "
        f"hardware_policy={args.hardware_policy}"
    )

    stream: Iterator[Any] = cli_api.generate_batch_iter(
        config,
        num_datasets=args.num_datasets,
        seed=run_seed,
        device=requested_device,
    )
    if diagnostics_aggregator is not None:
        base_stream = stream

        def _stream_with_diagnostics() -> Iterator[Any]:
            for bundle in base_stream:
                diagnostics_aggregator.update_bundle(bundle)
                yield bundle

        stream = _stream_with_diagnostics()

    if args.no_dataset_write:
        generated = sum(1 for _ in stream)
        if diagnostics_aggregator is not None:
            assert diagnostics_out_dir is not None
            _write_generate_diagnostics_artifacts(
                diagnostics_aggregator, diagnostics_out_dir=diagnostics_out_dir
            )
        print(f"Generated {generated} datasets (no-dataset-write mode).")
        return 0

    written = write_packed_parquet_shards_stream(
        stream,
        out_dir=out_dir,
        shard_size=config.output.shard_size,
        compression=config.output.compression,
    )
    if diagnostics_aggregator is not None:
        assert diagnostics_out_dir is not None
        _write_generate_diagnostics_artifacts(
            diagnostics_aggregator, diagnostics_out_dir=diagnostics_out_dir
        )
    print(f"Wrote {written} datasets to: {Path(out_dir)}")
    return 0



from __future__ import annotations

import json
from pathlib import Path
from types import SimpleNamespace

import pytest
import yaml
from conftest import load_repo_config, write_yaml

from dagzoo.cli import main
from dagzoo.cli.request_execution import run_request_execution
from dagzoo.config import (
    MISSINGNESS_MECHANISM_MCAR,
    REQUEST_FILE_VERSION_V1,
    REQUEST_PROFILE_DEFAULT,
    REQUEST_PROFILE_SMOKE,
    REQUEST_TASK_CLASSIFICATION,
    REQUEST_TASK_REGRESSION,
    GeneratorConfig,
    RequestFileConfig,
)
from dagzoo.config.io import load_packaged_generator_config
from dagzoo.core.config_resolution import resolve_request_config, serialize_resolution_events
from dagzoo.core.request_handoff import REQUEST_HANDOFF_SCHEMA_NAME
from dagzoo.filtering import DeferredFilterRunResult
from dagzoo.hardware import HardwareInfo


def _mock_cuda_h100(_requested_device: str) -> HardwareInfo:
    return HardwareInfo(
        backend="cuda",
        requested_device="cuda",
        device_name="NVIDIA H100 SXM",
        total_memory_gb=80.0,
        peak_flops=989e12,
        tier="cuda_h100",
    )


def _request_payload(**overrides: object) -> dict[str, object]:
    payload: dict[str, object] = {
        "version": REQUEST_FILE_VERSION_V1,
        "task": REQUEST_TASK_CLASSIFICATION,
        "dataset_count": 2,
        "rows": 1024,
        "profile": REQUEST_PROFILE_DEFAULT,
        "output_root": "requests/out",
    }
    payload.update(overrides)
    return payload


@pytest.mark.parametrize(
    "resource_name",
    [
        "default.yaml",
        "preset_missingness_mcar.yaml",
        "preset_missingness_mar.yaml",
        "preset_missingness_mnar.yaml",
    ],
)
def test_load_packaged_request_config_matches_repo_copy(resource_name: str) -> None:
    packaged = load_packaged_generator_config(resource_name)
    repo = load_repo_config(resource_name)

    assert packaged.to_dict() == repo.to_dict()


def test_resolve_request_config_loads_packaged_request_resources(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    loaded_resources: list[str] = []

    def _stub_load_packaged_generator_config(resource_name: str) -> GeneratorConfig:
        loaded_resources.append(resource_name)
        return load_repo_config(resource_name)

    monkeypatch.setattr(
        "dagzoo.core.config_resolution.load_packaged_generator_config",
        _stub_load_packaged_generator_config,
    )

    request = RequestFileConfig.from_dict(
        _request_payload(missingness_profile=MISSINGNESS_MECHANISM_MCAR)
    )

    resolved = resolve_request_config(
        request=request,
        device_override="cpu",
        hardware_policy="none",
    )

    assert loaded_resources == ["default.yaml", "preset_missingness_mcar.yaml"]
    assert resolved.config.dataset.missing_mechanism == MISSINGNESS_MECHANISM_MCAR


def test_resolve_request_config_maps_output_root_seed_and_rows() -> None:
    request = RequestFileConfig.from_dict(
        _request_payload(
            output_root="requests/demo_run",
            seed=123,
        )
    )

    resolved = resolve_request_config(
        request=request,
        device_override="cpu",
        hardware_policy="none",
    )

    assert resolved.config.seed == 123
    assert resolved.requested_device == "cpu"
    assert resolved.config.runtime.device == "cpu"
    assert resolved.config.output.out_dir == "requests/demo_run/generated"
    assert resolved.config.dataset.rows is not None
    assert resolved.config.dataset.rows.mode == "fixed"
    assert resolved.config.dataset.rows.value == 1024

    trace = serialize_resolution_events(resolved.trace_events)
    assert any(
        event["path"] == "seed" and event["source"] == "request.seed" and event["new_value"] == 123
        for event in trace
    )
    assert any(
        event["path"] == "output.out_dir"
        and event["source"] == "request.output_root"
        and event["new_value"] == "requests/demo_run/generated"
        for event in trace
    )


def test_resolve_request_config_applies_smoke_profile_without_overriding_rows() -> None:
    request = RequestFileConfig.from_dict(
        _request_payload(
            task=REQUEST_TASK_REGRESSION,
            profile=REQUEST_PROFILE_SMOKE,
            rows=1024,
        )
    )

    resolved = resolve_request_config(
        request=request,
        device_override="cpu",
        hardware_policy="none",
    )

    assert resolved.config.dataset.task == REQUEST_TASK_REGRESSION
    assert resolved.config.dataset.n_train == 128
    assert resolved.config.dataset.n_test == 32
    assert resolved.config.dataset.n_features_min == 8
    assert resolved.config.dataset.n_features_max == 12
    assert resolved.config.graph.n_nodes_min == 2
    assert resolved.config.graph.n_nodes_max == 12
    assert resolved.config.dataset.rows is not None
    assert resolved.config.dataset.rows.mode == "fixed"
    assert resolved.config.dataset.rows.value == 1024

    trace = serialize_resolution_events(resolved.trace_events)
    assert any(event["source"] == "request.profile_smoke" for event in trace)
    assert any(
        event["path"] == "dataset.rows" and event["source"] == "request.rows" for event in trace
    )


def test_resolve_request_config_preserves_smoke_caps_under_cuda_policy(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr("dagzoo.core.config_resolution.detect_hardware", _mock_cuda_h100)
    request = RequestFileConfig.from_dict(
        _request_payload(
            task=REQUEST_TASK_REGRESSION,
            profile=REQUEST_PROFILE_SMOKE,
            rows=1024,
        )
    )

    resolved = resolve_request_config(
        request=request,
        device_override="cuda",
        hardware_policy="cuda_tiered_v1",
    )

    assert resolved.requested_device == "cuda"
    assert resolved.config.dataset.n_train == 128
    assert resolved.config.dataset.n_test == 32
    assert resolved.config.dataset.n_features_max == 12
    assert resolved.config.graph.n_nodes_max == 12
    assert resolved.config.dataset.rows is not None
    assert resolved.config.dataset.rows.mode == "fixed"
    assert resolved.config.dataset.rows.value == 1024
    assert resolved.config.runtime.fixed_layout_target_cells == 160_000_000


def test_resolve_request_config_reapplies_non_smoke_rows_after_cuda_policy(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr("dagzoo.core.config_resolution.detect_hardware", _mock_cuda_h100)
    request = RequestFileConfig.from_dict(
        _request_payload(
            profile=REQUEST_PROFILE_DEFAULT,
            rows="2000..60000",
        )
    )

    resolved = resolve_request_config(
        request=request,
        device_override="cuda",
        hardware_policy="cuda_tiered_v1",
    )

    assert resolved.requested_device == "cuda"
    assert resolved.config.dataset.n_test == 1024
    assert resolved.config.dataset.rows is not None
    assert resolved.config.dataset.rows.mode == "range"
    assert resolved.config.dataset.rows.start == 2000
    assert resolved.config.dataset.rows.stop == 60000

    trace = serialize_resolution_events(resolved.trace_events)
    assert any(
        event["path"] == "dataset.rows"
        and event["source"] == "request.rows"
        and event["new_value"]
        == {
            "mode": "range",
            "value": None,
            "start": 2000,
            "stop": 60000,
            "choices": [],
        }
        for event in trace
    )


def test_resolve_request_config_applies_missingness_profile_without_task_leakage() -> None:
    request = RequestFileConfig.from_dict(
        _request_payload(
            task=REQUEST_TASK_REGRESSION,
            missingness_profile=MISSINGNESS_MECHANISM_MCAR,
        )
    )

    resolved = resolve_request_config(
        request=request,
        device_override="cpu",
        hardware_policy="none",
    )

    assert resolved.config.dataset.task == REQUEST_TASK_REGRESSION
    assert resolved.config.dataset.missing_rate == pytest.approx(0.2)
    assert resolved.config.dataset.missing_mechanism == MISSINGNESS_MECHANISM_MCAR
    trace = serialize_resolution_events(resolved.trace_events)
    assert any(
        event["path"] == "dataset.missing_mechanism"
        and event["source"] == "request.missingness_profile"
        and event["new_value"] == MISSINGNESS_MECHANISM_MCAR
        for event in trace
    )


def test_request_cli_end_to_end_writes_generated_filter_and_curated_outputs(
    tmp_path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    output_root = tmp_path / "request_run"
    request_path = write_yaml(
        tmp_path,
        "request.yaml",
        _request_payload(
            task=REQUEST_TASK_REGRESSION,
            dataset_count=1,
            rows=1024,
            profile=REQUEST_PROFILE_SMOKE,
            output_root=str(output_root),
        ),
    )

    def _stub_filter(*_args, **_kwargs):
        return True, {"wins_ratio": 1.0, "n_valid_oob": 128}

    monkeypatch.setattr(
        "dagzoo.filtering.deferred_filter._apply_extra_trees_filter_numpy", _stub_filter
    )

    code = main(
        [
            "request",
            "--request",
            str(request_path),
            "--device",
            "cpu",
            "--hardware-policy",
            "none",
        ]
    )

    assert code == 0
    effective_config = yaml.safe_load(
        (output_root / "generated" / "effective_config.yaml").read_text(encoding="utf-8")
    )
    trace_payload = yaml.safe_load(
        (output_root / "generated" / "effective_config_trace.yaml").read_text(encoding="utf-8")
    )
    assert effective_config["output"]["out_dir"] == str(output_root / "generated")
    assert effective_config["dataset"]["rows"] is None
    assert int(effective_config["dataset"]["n_train"]) == 128
    assert int(effective_config["dataset"]["n_test"]) == 32
    assert any(
        isinstance(item, dict) and item.get("source") == "request.smoke_rows_cap"
        for item in trace_payload
    )
    assert (output_root / "generated" / "shard_00000" / "metadata.ndjson").exists()
    assert (output_root / "filter" / "filter_manifest.ndjson").exists()
    summary = json.loads((output_root / "filter" / "filter_summary.json").read_text("utf-8"))
    assert summary["total_datasets"] == 1
    assert summary["accepted_datasets"] == 1
    assert summary["rejected_datasets"] == 0
    assert (output_root / "curated" / "shard_00000" / "metadata.ndjson").exists()
    handoff = json.loads((output_root / "handoff_manifest.json").read_text(encoding="utf-8"))
    assert handoff["schema_name"] == REQUEST_HANDOFF_SCHEMA_NAME
    assert handoff["artifacts"]["filtered_corpus_dir"] == str((output_root / "curated").resolve())
    assert handoff["artifacts"]["filter_summary_path"] == str(
        (output_root / "filter" / "filter_summary.json").resolve()
    )
    assert handoff["summary"]["accepted_datasets"] == 1
    assert handoff["diversity_artifacts"]["summary_json_path"] is None


def test_request_execution_manifest_uses_wall_clock_filter_timing(
    tmp_path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    output_root = tmp_path / "request_run"
    request_path = write_yaml(
        tmp_path,
        "request.yaml",
        _request_payload(
            task=REQUEST_TASK_REGRESSION,
            dataset_count=2,
            rows=1024,
            profile=REQUEST_PROFILE_DEFAULT,
            output_root=str(output_root),
        ),
    )

    perf_counter_values = iter((10.0, 16.0, 20.0, 50.0))
    monkeypatch.setattr(
        "dagzoo.cli.request_execution.perf_counter", lambda: next(perf_counter_values)
    )

    def _stub_write_packed_parquet_shards_stream(
        _stream,
        *,
        out_dir: Path,
        shard_size: int,
        compression: str,
    ) -> int:
        assert shard_size > 0
        assert compression
        out_dir.mkdir(parents=True, exist_ok=True)
        return 2

    def _stub_run_deferred_filter(**kwargs) -> DeferredFilterRunResult:
        assert kwargs["in_dir"] == output_root / "generated"
        assert kwargs["out_dir"] == output_root / "filter"
        assert kwargs["curated_out_dir"] == output_root / "curated"
        return DeferredFilterRunResult(
            manifest_path=output_root / "filter" / "filter_manifest.ndjson",
            summary_path=output_root / "filter" / "filter_summary.json",
            total_datasets=2,
            accepted_datasets=1,
            rejected_datasets=1,
            elapsed_seconds=0.5,
            datasets_per_minute=999.0,
            curated_out_dir=output_root / "curated",
            curated_accepted_datasets=1,
        )

    monkeypatch.setattr(
        "dagzoo.cli.request_execution.write_packed_parquet_shards_stream",
        _stub_write_packed_parquet_shards_stream,
    )
    monkeypatch.setattr(
        "dagzoo.cli.request_execution.get_cli_public_api",
        lambda: SimpleNamespace(
            generate_batch_iter=lambda *_args, **_kwargs: iter(()),
            run_deferred_filter=_stub_run_deferred_filter,
        ),
    )

    _ = run_request_execution(
        request_path=request_path,
        device_override="cpu",
        hardware_policy="none",
        n_jobs_override=None,
        print_effective_config_flag=False,
        print_resolution_trace_flag=False,
    )

    handoff = json.loads((output_root / "handoff_manifest.json").read_text(encoding="utf-8"))
    assert handoff["throughput"]["generation_stage"]["elapsed_seconds"] == pytest.approx(6.0)
    assert handoff["throughput"]["generation_stage"]["datasets_per_minute"] == pytest.approx(20.0)
    assert handoff["throughput"]["filter_stage"]["elapsed_seconds"] == pytest.approx(30.0)
    assert handoff["throughput"]["filter_stage"]["datasets_per_minute"] == pytest.approx(4.0)



"""Tests for graph/dag_sampler.py."""

import pytest
import torch
from conftest import make_generator as _make_generator

from dagzoo.graph.dag_sampler import (
    dag_edge_density,
    dag_longest_path_nodes,
    sample_dag,
)


def test_dag_shape() -> None:
    g = _make_generator(42)
    adj = sample_dag(5, g)
    assert adj.shape == (5, 5)
    assert adj.dtype == torch.bool


def test_dag_upper_triangular() -> None:
    g = _make_generator(7)
    adj = sample_dag(8, g)
    for i in range(8):
        for j in range(i + 1):
            assert adj[i, j] == False  # noqa: E712


def test_dag_deterministic() -> None:
    a = sample_dag(6, _make_generator(99))
    b = sample_dag(6, _make_generator(99))
    torch.testing.assert_close(a, b)


def test_dag_deterministic_with_edge_logit_bias() -> None:
    a = sample_dag(6, _make_generator(99), edge_logit_bias=0.75)
    b = sample_dag(6, _make_generator(99), edge_logit_bias=0.75)
    torch.testing.assert_close(a, b)


def test_dag_single_node() -> None:
    g = _make_generator(0)
    adj = sample_dag(1, g)
    assert adj.shape == (1, 1)
    assert not adj[0, 0]


def test_edge_logit_bias_increases_edge_count_with_same_rng_stream() -> None:
    low_bias = sample_dag(16, _make_generator(123), edge_logit_bias=-1.0)
    high_bias = sample_dag(16, _make_generator(123), edge_logit_bias=1.0)
    assert int(high_bias.sum().item()) >= int(low_bias.sum().item())


def test_dag_longest_path_nodes_on_known_graph() -> None:
    adjacency = torch.tensor(
        [
            [0, 1, 1, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1],
            [0, 0, 0, 0],
        ],
        dtype=torch.bool,
    )
    assert dag_longest_path_nodes(adjacency) == 4


def test_dag_longest_path_nodes_rejects_non_upper_triangular_input() -> None:
    adjacency = torch.tensor(
        [
            [0, 1, 0],
            [1, 0, 0],
            [0, 0, 0],
        ],
        dtype=torch.bool,
    )
    with pytest.raises(ValueError, match="upper-triangular"):
        dag_longest_path_nodes(adjacency)


def test_dag_longest_path_nodes_rejects_singleton_self_loop() -> None:
    adjacency = torch.tensor([[1]], dtype=torch.bool)
    with pytest.raises(ValueError, match="upper-triangular"):
        dag_longest_path_nodes(adjacency)


def test_dag_edge_density_on_known_graph() -> None:
    adjacency = torch.tensor(
        [
            [0, 1, 1, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
        ],
        dtype=torch.bool,
    )
    assert dag_edge_density(adjacency) == 0.5


def test_dag_edge_density_single_node_is_zero() -> None:
    adjacency = torch.zeros((1, 1), dtype=torch.bool)
    assert dag_edge_density(adjacency) == 0.0


def test_dag_edge_density_rejects_singleton_self_loop() -> None:
    adjacency = torch.tensor([[1]], dtype=torch.bool)
    with pytest.raises(ValueError, match="upper-triangular"):
        dag_edge_density(adjacency)



import json

import numpy as np
import pytest

from dagzoo.filtering.deferred_filter import _iter_packed_split_datasets, run_deferred_filter
from dagzoo.io.parquet_writer import write_packed_parquet_shards_stream
from dagzoo.types import DatasetBundle


def _bundle_with_embedded_config(
    seed: int,
    *,
    dataset_seed: int | None = None,
    dataset_index: int | None = None,
) -> DatasetBundle:
    metadata = {
        "seed": seed,
        "filter": {"mode": "deferred", "status": "not_run"},
        "config": {
            "dataset": {"task": "classification"},
            "filter": {"enabled": True},
        },
    }
    if dataset_seed is not None:
        metadata["dataset_seed"] = int(dataset_seed)
    if dataset_index is not None:
        metadata["dataset_index"] = int(dataset_index)

    return DatasetBundle(
        X_train=np.array(
            [
                [0.0, 1.0],
                [1.0, 0.0],
                [2.0, 1.0],
            ],
            dtype=np.float32,
        ),
        y_train=np.array([0, 1, 0], dtype=np.int64),
        X_test=np.array([[1.5, 0.5], [0.5, 1.5]], dtype=np.float32),
        y_test=np.array([1, 0], dtype=np.int64),
        feature_types=["num", "num"],
        metadata=metadata,
    )


def _bundle_without_config(seed: int) -> DatasetBundle:
    bundle = _bundle_with_embedded_config(seed)
    bundle.metadata["config"] = {"dataset": {"task": "classification"}}
    return bundle


def _load_ndjson(path) -> list[dict[str, object]]:
    payload: list[dict[str, object]] = []
    for line in path.read_text(encoding="utf-8").splitlines():
        if line.strip():
            payload.append(json.loads(line))
    return payload


def _write_ndjson_records(path, records: list[dict[str, object]]) -> None:
    path.write_text(
        "".join(json.dumps(record, sort_keys=True) + "\n" for record in records),
        encoding="utf-8",
    )


def _write_split_table(
    path,
    *,
    dataset_indices: list[int],
    row_indices: list[int],
    x_rows: list[list[float]],
    y_rows: list[int],
) -> None:
    pyarrow = pytest.importorskip("pyarrow")
    pyarrow_parquet = pytest.importorskip("pyarrow.parquet")

    table = pyarrow.table(
        {
            "dataset_index": pyarrow.array(dataset_indices, type=pyarrow.int64()),
            "row_index": pyarrow.array(row_indices, type=pyarrow.int64()),
            "x": pyarrow.array(x_rows, type=pyarrow.list_(pyarrow.float32())),
            "y": pyarrow.array(y_rows, type=pyarrow.int64()),
        }
    )
    pyarrow_parquet.write_table(table, path, compression="zstd")


def test_iter_packed_split_datasets_handles_dataset_split_across_record_batches(
    tmp_path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    pyarrow = pytest.importorskip("pyarrow")

    table = pyarrow.table(
        {
            "dataset_index": pyarrow.array([0, 0, 0, 1, 1], type=pyarrow.int64()),
            "row_index": pyarrow.array([0, 1, 2, 0, 1], type=pyarrow.int64()),
            "x": pyarrow.array(
                [[0.0, 0.5], [1.0, 1.5], [2.0, 2.5], [3.0, 3.5], [4.0, 4.5]],
                type=pyarrow.list_(pyarrow.float32()),
            ),
            "y": pyarrow.array([0, 1, 0, 1, 0], type=pyarrow.int64()),
        }
    )
    batches = [
        table.slice(0, 2).to_batches()[0],
        table.slice(2, 1).to_batches()[0],
        table.slice(3, 2).to_batches()[0],
    ]

    class _FakeParquetFile:
        def __init__(self, _path) -> None:
            self.schema_arrow = table.schema

        def iter_batches(self, *, columns):
            assert columns == ["dataset_index", "row_index", "x", "y"]
            return iter(batches)

    monkeypatch.setattr("dagzoo.filtering.deferred_filter.pq.ParquetFile", _FakeParquetFile)

    split_path = tmp_path / "train.parquet"
    datasets = list(_iter_packed_split_datasets(split_path))

    assert [dataset.dataset_index for dataset in datasets] == [0, 1]
    assert datasets[0].x.shape == (3, 2)
    assert datasets[0].y.tolist() == [0, 1, 0]
    assert np.allclose(datasets[0].x[:, 0], np.array([0.0, 1.0, 2.0], dtype=np.float32))
    assert datasets[1].x.shape == (2, 2)
    assert datasets[1].y.tolist() == [1, 0]


def test_iter_packed_split_datasets_handles_mixed_feature_widths_within_record_batch(
    tmp_path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    pyarrow = pytest.importorskip("pyarrow")

    table = pyarrow.table(
        {
            "dataset_index": pyarrow.array([0, 0, 1, 1], type=pyarrow.int64()),
            "row_index": pyarrow.array([0, 1, 0, 1], type=pyarrow.int64()),
            "x": pyarrow.array(
                [[0.0, 0.5], [1.0, 1.5], [2.0, 2.5, 2.75], [3.0, 3.5, 3.75]],
                type=pyarrow.list_(pyarrow.float32()),
            ),
            "y": pyarrow.array([0, 1, 1, 0], type=pyarrow.int64()),
        }
    )
    batches = table.to_batches(max_chunksize=4)

    class _FakeParquetFile:
        def __init__(self, _path) -> None:
            self.schema_arrow = table.schema

        def iter_batches(self, *, columns):
            assert columns == ["dataset_index", "row_index", "x", "y"]
            return iter(batches)

    monkeypatch.setattr("dagzoo.filtering.deferred_filter.pq.ParquetFile", _FakeParquetFile)

    split_path = tmp_path / "train.parquet"
    datasets = list(_iter_packed_split_datasets(split_path))

    assert [dataset.dataset_index for dataset in datasets] == [0, 1]
    assert datasets[0].x.shape == (2, 2)
    assert datasets[0].y.tolist() == [0, 1]
    assert datasets[1].x.shape == (2, 3)
    assert datasets[1].y.tolist() == [1, 0]


def test_run_deferred_filter_writes_manifest_and_summary(
    tmp_path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    pytest.importorskip("pyarrow.parquet")

    in_dir = tmp_path / "input"
    out_dir = tmp_path / "filter_out"
    bundles = [_bundle_with_embedded_config(101), _bundle_with_embedded_config(102)]
    _ = write_packed_parquet_shards_stream(bundles, in_dir, shard_size=2, compression="zstd")

    def _stub_filter(*_args, **_kwargs):
        seed = int(_kwargs["seed"])
        accepted = bool(seed % 2)
        details = {"wins_ratio": 1.0 if accepted else 0.0, "n_valid_oob": 128}
        if not accepted:
            details["reason"] = "below_threshold"
        return accepted, details

    monkeypatch.setattr(
        "dagzoo.filtering.deferred_filter._apply_extra_trees_filter_numpy", _stub_filter
    )

    result = run_deferred_filter(in_dir=in_dir, out_dir=out_dir)
    assert result.total_datasets == 2
    assert result.accepted_datasets == 1
    assert result.rejected_datasets == 1

    summary = json.loads(result.summary_path.read_text(encoding="utf-8"))
    assert summary["accepted_datasets"] == 1
    assert summary["rejected_datasets"] == 1
    assert summary["rejected_reason_counts"]["below_threshold"] == 1

    manifest_records = _load_ndjson(result.manifest_path)
    assert len(manifest_records) == 2
    statuses = {str(record["status"]) for record in manifest_records}
    assert statuses == {"accepted", "rejected"}
    for record in manifest_records:
        filter_payload = record["filter"]
        assert isinstance(filter_payload, dict)
        assert filter_payload["mode"] == "deferred"
        assert filter_payload["status"] in {"accepted", "rejected"}


def test_run_deferred_filter_writes_curated_output_for_accepted_only(
    tmp_path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    pyarrow_parquet = pytest.importorskip("pyarrow.parquet")

    in_dir = tmp_path / "input"
    out_dir = tmp_path / "filter_out"
    curated_out = tmp_path / "curated"
    bundles = [
        _bundle_with_embedded_config(201),
        _bundle_with_embedded_config(202),
        _bundle_with_embedded_config(203),
    ]
    _ = write_packed_parquet_shards_stream(bundles, in_dir, shard_size=3, compression="zstd")

    def _stub_filter(*_args, **_kwargs):
        seed = int(_kwargs["seed"])
        accepted = bool(seed % 2)
        return accepted, {"wins_ratio": 1.0 if accepted else 0.0, "n_valid_oob": 128}

    monkeypatch.setattr(
        "dagzoo.filtering.deferred_filter._apply_extra_trees_filter_numpy", _stub_filter
    )

    result = run_deferred_filter(in_dir=in_dir, out_dir=out_dir, curated_out_dir=curated_out)
    assert result.curated_accepted_datasets == 2

    shard_dir = curated_out / "shard_00000"
    assert shard_dir.exists()
    metadata_records = _load_ndjson(shard_dir / "metadata.ndjson")
    assert [int(record["dataset_index"]) for record in metadata_records] == [0, 2]

    train_table = pyarrow_parquet.read_table(shard_dir / "train.parquet")
    dataset_indices = {int(value) for value in train_table.column("dataset_index").to_pylist()}
    assert dataset_indices == {0, 2}


def test_run_deferred_filter_requires_embedded_filter_config(
    tmp_path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    pytest.importorskip("pyarrow.parquet")

    in_dir = tmp_path / "input"
    out_dir = tmp_path / "filter_out"
    bundles = [_bundle_without_config(7)]
    _ = write_packed_parquet_shards_stream(bundles, in_dir, shard_size=1, compression="zstd")

    monkeypatch.setattr(
        "dagzoo.filtering.deferred_filter._apply_extra_trees_filter_numpy",
        lambda *_args, **_kwargs: (True, {"wins_ratio": 1.0, "n_valid_oob": 128}),
    )

    with pytest.raises(ValueError, match="requires embedded metadata\\.config\\.filter"):
        _ = run_deferred_filter(in_dir=in_dir, out_dir=out_dir)


def test_run_deferred_filter_prefers_dataset_seed_when_present(
    tmp_path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    pytest.importorskip("pyarrow.parquet")

    in_dir = tmp_path / "input"
    out_dir = tmp_path / "filter_out"
    bundles = [
        _bundle_with_embedded_config(101, dataset_seed=501, dataset_index=0),
        _bundle_with_embedded_config(102, dataset_seed=502, dataset_index=1),
    ]
    _ = write_packed_parquet_shards_stream(bundles, in_dir, shard_size=2, compression="zstd")

    replay_seeds: list[int] = []

    def _stub_filter(*_args, **_kwargs):
        replay_seeds.append(int(_kwargs["seed"]))
        return True, {"wins_ratio": 1.0, "n_valid_oob": 128}

    monkeypatch.setattr(
        "dagzoo.filtering.deferred_filter._apply_extra_trees_filter_numpy", _stub_filter
    )

    result = run_deferred_filter(in_dir=in_dir, out_dir=out_dir)

    assert result.total_datasets == 2
    assert result.accepted_datasets == 2
    assert replay_seeds == [501, 502]


def test_run_deferred_filter_rejects_stale_filter_output_dir(
    tmp_path,
) -> None:
    pytest.importorskip("pyarrow.parquet")

    in_dir = tmp_path / "input"
    out_dir = tmp_path / "filter_out"
    _ = write_packed_parquet_shards_stream(
        [_bundle_with_embedded_config(301)],
        in_dir,
        shard_size=1,
        compression="zstd",
    )

    out_dir.mkdir()
    (out_dir / "filter_manifest.ndjson").write_text("", encoding="utf-8")

    with pytest.raises(RuntimeError, match="already contains prior artifacts"):
        _ = run_deferred_filter(in_dir=in_dir, out_dir=out_dir)


def test_run_deferred_filter_rejects_extra_split_rows_beyond_metadata(
    tmp_path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    pytest.importorskip("pyarrow.parquet")

    in_dir = tmp_path / "input"
    out_dir = tmp_path / "filter_out"
    _ = write_packed_parquet_shards_stream(
        [_bundle_with_embedded_config(401), _bundle_with_embedded_config(402)],
        in_dir,
        shard_size=2,
        compression="zstd",
    )

    metadata_path = in_dir / "shard_00000" / "metadata.ndjson"
    records = _load_ndjson(metadata_path)
    _write_ndjson_records(metadata_path, [records[0]])

    monkeypatch.setattr(
        "dagzoo.filtering.deferred_filter._apply_extra_trees_filter_numpy",
        lambda *_args, **_kwargs: (True, {"wins_ratio": 1.0, "n_valid_oob": 128}),
    )

    with pytest.raises(ValueError, match="extra dataset rows beyond metadata coverage"):
        _ = run_deferred_filter(in_dir=in_dir, out_dir=out_dir)
    assert list(out_dir.iterdir()) == []

    with pytest.raises(ValueError, match="extra dataset rows beyond metadata coverage"):
        _ = run_deferred_filter(in_dir=in_dir, out_dir=out_dir)
    assert list(out_dir.iterdir()) == []


def test_run_deferred_filter_rejects_non_monotonic_split_rows(
    tmp_path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    pytest.importorskip("pyarrow.parquet")

    shard_dir = tmp_path / "input" / "shard_00000"
    shard_dir.mkdir(parents=True)
    metadata_path = shard_dir / "metadata.ndjson"
    train_path = shard_dir / "train.parquet"
    test_path = shard_dir / "test.parquet"

    metadata_records = [
        {
            "dataset_index": 0,
            "n_train": 2,
            "n_test": 1,
            "n_features": 2,
            "feature_types": ["num", "num"],
            "metadata": {
                "seed": 11,
                "filter": {"mode": "deferred", "status": "not_run"},
                "config": {
                    "dataset": {"task": "classification"},
                    "filter": {"enabled": True},
                },
            },
        },
        {
            "dataset_index": 1,
            "n_train": 1,
            "n_test": 1,
            "n_features": 2,
            "feature_types": ["num", "num"],
            "metadata": {
                "seed": 12,
                "filter": {"mode": "deferred", "status": "not_run"},
                "config": {
                    "dataset": {"task": "classification"},
                    "filter": {"enabled": True},
                },
            },
        },
    ]
    _write_ndjson_records(metadata_path, metadata_records)
    _write_split_table(
        train_path,
        dataset_indices=[0, 1, 0],
        row_indices=[0, 0, 1],
        x_rows=[[0.0, 0.0], [1.0, 1.0], [0.5, 0.5]],
        y_rows=[0, 1, 0],
    )
    _write_split_table(
        test_path,
        dataset_indices=[0, 1],
        row_indices=[0, 0],
        x_rows=[[0.25, 0.25], [1.25, 1.25]],
        y_rows=[0, 1],
    )

    monkeypatch.setattr(
        "dagzoo.filtering.deferred_filter._apply_extra_trees_filter_numpy",
        lambda *_args, **_kwargs: (True, {"wins_ratio": 1.0, "n_valid_oob": 128}),
    )

    with pytest.raises(ValueError, match="monotonically increasing dataset_index"):
        _ = run_deferred_filter(in_dir=shard_dir, out_dir=tmp_path / "filter_out")


def test_run_deferred_filter_rejects_lineage_symlinks_during_curated_copy(
    tmp_path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    pytest.importorskip("pyarrow.parquet")

    in_dir = tmp_path / "input"
    out_dir = tmp_path / "filter_out"
    curated_out = tmp_path / "curated_out"
    _ = write_packed_parquet_shards_stream(
        [_bundle_with_embedded_config(501)],
        in_dir,
        shard_size=1,
        compression="zstd",
    )

    lineage_dir = in_dir / "shard_00000" / "lineage"
    lineage_dir.mkdir()
    outside_path = tmp_path / "outside.txt"
    outside_path.write_text("sentinel\n", encoding="utf-8")
    try:
        (lineage_dir / "escape.txt").symlink_to(outside_path)
    except OSError:
        pytest.skip("symlinks unavailable in this environment")

    monkeypatch.setattr(
        "dagzoo.filtering.deferred_filter._apply_extra_trees_filter_numpy",
        lambda *_args, **_kwargs: (True, {"wins_ratio": 1.0, "n_valid_oob": 128}),
    )

    with pytest.raises(RuntimeError, match="must not be a symlink"):
        _ = run_deferred_filter(in_dir=in_dir, out_dir=out_dir, curated_out_dir=curated_out)
    assert list(out_dir.iterdir()) == []
    assert list(curated_out.iterdir()) == []

    with pytest.raises(RuntimeError, match="must not be a symlink"):
        _ = run_deferred_filter(in_dir=in_dir, out_dir=out_dir, curated_out_dir=curated_out)
    assert list(out_dir.iterdir()) == []
    assert list(curated_out.iterdir()) == []


def test_run_deferred_filter_cleans_up_curated_output_after_split_exhaustion_failure(
    tmp_path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    pytest.importorskip("pyarrow.parquet")

    in_dir = tmp_path / "input"
    out_dir = tmp_path / "filter_out"
    curated_out = tmp_path / "curated_out"
    _ = write_packed_parquet_shards_stream(
        [_bundle_with_embedded_config(601), _bundle_with_embedded_config(602)],
        in_dir,
        shard_size=2,
        compression="zstd",
    )

    metadata_path = in_dir / "shard_00000" / "metadata.ndjson"
    records = _load_ndjson(metadata_path)
    _write_ndjson_records(metadata_path, [records[0]])

    monkeypatch.setattr(
        "dagzoo.filtering.deferred_filter._apply_extra_trees_filter_numpy",
        lambda *_args, **_kwargs: (True, {"wins_ratio": 1.0, "n_valid_oob": 128}),
    )

    with pytest.raises(ValueError, match="extra dataset rows beyond metadata coverage"):
        _ = run_deferred_filter(in_dir=in_dir, out_dir=out_dir, curated_out_dir=curated_out)
    assert list(out_dir.iterdir()) == []
    assert list(curated_out.iterdir()) == []

    with pytest.raises(ValueError, match="extra dataset rows beyond metadata coverage"):
        _ = run_deferred_filter(in_dir=in_dir, out_dir=out_dir, curated_out_dir=curated_out)
    assert list(out_dir.iterdir()) == []
    assert list(curated_out.iterdir()) == []
